# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=unsubscriptable-object

"""
Implements various generic attrs validators.
"""

import warnings
from enum import Enum
from typing import Any, List, Type

import attr
from attr import Attribute, attrs


@attrs(repr=False, slots=True, hash=True)
class _EnumValidator:
    """Validator for use by enum(), following the pattern from the standard attrs _InValidator."""

    options = attr.ib(type=Type[Enum])

    def __call__(self, instance: Any, attribute: Attribute, value: Enum) -> None:  # type: ignore
        try:
            # This gives us "DeprecationWarning: using non-Enums in containment checks will raise TypeError in Python 3.8"
            # Uh, yeah.  I know.  I'm catching TypeError.  And anyway, this method is restricted to enums.
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=DeprecationWarning)
                in_options = value in self.options
        except TypeError:
            in_options = False
        if not in_options:
            legal = ", ".join(sorted([option.name for option in list(self.options)]))  # type: ignore
            raise ValueError("'%s' must be one of [%s]" % (attribute.name, legal))


@attrs(repr=False, slots=True, hash=True)
class _LengthValidator:
    """Validator for use by maxlength(), following the pattern from the standard attrs _InValidator."""

    maxlength = attr.ib(type=int)

    def __call__(self, instance: Any, attribute: Attribute, value: str) -> None:  # type: ignore
        if len(value) > self.maxlength:
            raise ValueError("'%s' must not exceed length %d" % (attribute.name, self.maxlength))


def enum(options: Type[Enum]) -> _EnumValidator:
    """attrs validator to ensure that a value is a legal enumeration."""
    return _EnumValidator(options)


def length(maxlength: int) -> _LengthValidator:
    """attrs validator to ensure that a string value does not exceed a length."""
    return _LengthValidator(maxlength)


def notempty(_instance: Any, attribute: Attribute, value: Any) -> None:  # type: ignore
    """attrs validator to ensure that a list is not empty."""
    if len(value) == 0:
        raise ValueError("'%s' may not be empty" % attribute.name)


def string(_instance: Any, attribute: Attribute, value: str) -> None:  # type: ignore
    """attrs validator to ensure that a string is non-empty."""
    # Annoyingly, due to some quirk in the CattrConverter, we end up with "None" rather than None for strings set to JSON null
    # As a result, we need to prevent "None" as a legal value, but that's probably better anyway.
    if value is None or value == "None" or not isinstance(value, str) or len(value) == 0:
        raise ValueError("'%s' must be a non-empty string" % attribute.name)


def stringlist(_instance: Any, attribute: Attribute, value: List[str]) -> None:  # type: ignore
    """attrs validator to ensure that a string list contains non-empty values."""
    # Annoyingly, due to some quirk in the CattrConverter, we end up with "None" rather than None for strings set to JSON null
    # As a result, we need to prevent "None" as a legal value, but that's probably better anyway.
    for element in value:
        if element is None or element == "None" or not isinstance(element, str) or len(element) == 0:
            raise ValueError("'%s' elements must be non-empty strings" % attribute.name)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
System configuration.
"""

import configparser
import os
from configparser import ConfigParser, SectionProxy
from typing import Any, Dict, Optional, Union

import attr
import cattr
import orjson

from .util import homedir

# Configuration defaults
DEFAULT_CONFIG_PATH = os.path.join(homedir(), ".apologiesrc")
DEFAULT_LOGFILE_PATH = None
DEFAULT_SERVER_HOST = "localhost"
DEFAULT_SERVER_PORT = 8080
DEFAULT_CLOSE_TIMEOUT_SEC = 10
DEFAULT_WEBSOCKET_LIMIT = 1000
DEFAULT_TOTAL_GAME_LIMIT = 1000
DEFAULT_IN_PROGRESS_GAME_LIMIT = 25
DEFAULT_REGISTERED_PLAYER_LIMIT = 100
DEFAULT_WEBSOCKET_IDLE_THRESH_MIN = 2
DEFAULT_WEBSOCKET_INACTIVE_THRESH_MIN = 5
DEFAULT_PLAYER_IDLE_THRESH_MIN = 15
DEFAULT_PLAYER_INACTIVE_THRESH_MIN = 30
DEFAULT_GAME_IDLE_THRESH_MIN = 10
DEFAULT_GAME_INACTIVE_THRESH_MIN = 20
DEFAULT_GAME_RETENTION_THRESH_MIN = 2880  # 2 days
IDLE_WEBSOCKET_CHECK_PERIOD_SEC = 120
IDLE_WEBSOCKET_CHECK_DELAY_SEC = 300
IDLE_PLAYER_CHECK_PERIOD_SEC = 120
IDLE_PLAYER_CHECK_DELAY_SEC = 300
IDLE_GAME_CHECK_PERIOD_SEC = 120
IDLE_GAME_CHECK_DELAY_SEC = 300
OBSOLETE_GAME_CHECK_PERIOD_SEC = 300
OBSOLETE_GAME_CHECK_DELAY_SEC = 300

DEFAULTS = {
    "logfile_path": DEFAULT_LOGFILE_PATH,
    "server_host": DEFAULT_SERVER_HOST,
    "server_port": DEFAULT_SERVER_PORT,
    "close_timeout_sec": DEFAULT_CLOSE_TIMEOUT_SEC,
    "websocket_limit": DEFAULT_WEBSOCKET_LIMIT,
    "total_game_limit": DEFAULT_TOTAL_GAME_LIMIT,
    "in_progress_game_limit": DEFAULT_IN_PROGRESS_GAME_LIMIT,
    "registered_player_limit": DEFAULT_REGISTERED_PLAYER_LIMIT,
    "websocket_idle_thresh_min": DEFAULT_WEBSOCKET_IDLE_THRESH_MIN,
    "websocket_inactive_thresh_min": DEFAULT_WEBSOCKET_INACTIVE_THRESH_MIN,
    "player_idle_thresh_min": DEFAULT_PLAYER_IDLE_THRESH_MIN,
    "player_inactive_thresh_min": DEFAULT_PLAYER_INACTIVE_THRESH_MIN,
    "game_idle_thresh_min": DEFAULT_GAME_IDLE_THRESH_MIN,
    "game_inactive_thresh_min": DEFAULT_GAME_INACTIVE_THRESH_MIN,
    "game_retention_thresh_min": DEFAULT_GAME_RETENTION_THRESH_MIN,
    "idle_websocket_check_period_sec": IDLE_WEBSOCKET_CHECK_PERIOD_SEC,
    "idle_websocket_check_delay_sec": IDLE_WEBSOCKET_CHECK_DELAY_SEC,
    "idle_player_check_period_sec": IDLE_PLAYER_CHECK_PERIOD_SEC,
    "idle_player_check_delay_sec": IDLE_PLAYER_CHECK_DELAY_SEC,
    "idle_game_check_period_sec": IDLE_GAME_CHECK_PERIOD_SEC,
    "idle_game_check_delay_sec": IDLE_GAME_CHECK_DELAY_SEC,
    "obsolete_game_check_period_sec": OBSOLETE_GAME_CHECK_PERIOD_SEC,
    "obsolete_game_check_delay_sec": OBSOLETE_GAME_CHECK_DELAY_SEC,
}

# pylint: disable=too-many-instance-attributes
@attr.s(frozen=True)
class SystemConfig:
    """
    System configuration.
    
    Attributes:
        logfile_path(str): The path to the log file on disk
        server_host(str): The hostname to bind to
        server_port(int): The server port to listen on
        close_timeout_sec(int): The maximum amount of time to wait when closing a connection
        websocket_limit(int): Limit on the number of websocket connections
        total_game_limit(int): Limit on the total number of tracked games
        in_progress_game_limit(int): Limit on the number of in progress games
        registered_player_limit(int): Limit on the number of registered players
        websocket_idle_thresh_min(int): Number of minutes of no activity before a websocket with no players is considered idle
        websocket_inactive_thresh_min(int): Number of minutes of no activity before a websocket with no players is inactive
        player_idle_thresh_min(int): Number of minutes of no activity before a player is considered idle
        player_inactive_thresh_min(int): Number of minutes of no activity before a player is considered inactive
        game_idle_thresh_min(int): Number of minutes of no activity before a game is considered idle
        game_inactive_thresh_min(int): Number of minutes of no activity before a game is considered inactive
        game_retention_thresh_min(int): Number of minutes to retain data about games after they are completed
        idle_websocket_check_period_sec(int): Number of seconds to wait between executions of the Idle Websocket Check task
        idle_websocket_check_delay_sec(int): Number of seconds to delay before the first Idle Websocket Check task
        idle_player_check_period_sec(int): Number of seconds to wait between executions of the Idle Player Check task
        idle_player_check_delay_sec(int): Number of seconds to delay before the first Idle Player Check task
        idle_game_check_period_sec(int): Number of seconds to wait between executions of the Idle Game Check task
        idle_game_check_delay_sec(int): Number of seconds to delay before the first Idle Player Check task
        obsolete_game_check_period_sec(int):  Number of seconds to wait between executions of the Obsolete Game Check task
        obsolete_game_check_delay_sec(int): Number of seconds to delay before the first Idle Player Check task
    """

    logfile_path = attr.ib(type=str, default=DEFAULT_LOGFILE_PATH)
    server_host = attr.ib(type=str, default=DEFAULT_SERVER_HOST)
    server_port = attr.ib(type=int, default=DEFAULT_SERVER_PORT)
    close_timeout_sec = attr.ib(type=int, default=DEFAULT_CLOSE_TIMEOUT_SEC)
    websocket_limit = attr.ib(type=int, default=DEFAULT_WEBSOCKET_LIMIT)
    total_game_limit = attr.ib(type=int, default=DEFAULT_TOTAL_GAME_LIMIT)
    in_progress_game_limit = attr.ib(type=int, default=DEFAULT_IN_PROGRESS_GAME_LIMIT)
    registered_player_limit = attr.ib(type=int, default=DEFAULT_REGISTERED_PLAYER_LIMIT)
    websocket_idle_thresh_min = attr.ib(type=int, default=DEFAULT_WEBSOCKET_IDLE_THRESH_MIN)
    websocket_inactive_thresh_min = attr.ib(type=int, default=DEFAULT_WEBSOCKET_INACTIVE_THRESH_MIN)
    player_idle_thresh_min = attr.ib(type=int, default=DEFAULT_PLAYER_IDLE_THRESH_MIN)
    player_inactive_thresh_min = attr.ib(type=int, default=DEFAULT_PLAYER_INACTIVE_THRESH_MIN)
    game_idle_thresh_min = attr.ib(type=int, default=DEFAULT_GAME_IDLE_THRESH_MIN)
    game_inactive_thresh_min = attr.ib(type=int, default=DEFAULT_GAME_INACTIVE_THRESH_MIN)
    game_retention_thresh_min = attr.ib(type=int, default=DEFAULT_GAME_RETENTION_THRESH_MIN)
    idle_websocket_check_period_sec = attr.ib(type=int, default=IDLE_WEBSOCKET_CHECK_PERIOD_SEC)
    idle_websocket_check_delay_sec = attr.ib(type=int, default=IDLE_WEBSOCKET_CHECK_DELAY_SEC)
    idle_player_check_period_sec = attr.ib(type=int, default=IDLE_PLAYER_CHECK_PERIOD_SEC)
    idle_player_check_delay_sec = attr.ib(type=int, default=IDLE_PLAYER_CHECK_DELAY_SEC)
    idle_game_check_period_sec = attr.ib(type=int, default=IDLE_GAME_CHECK_PERIOD_SEC)
    idle_game_check_delay_sec = attr.ib(type=int, default=IDLE_GAME_CHECK_DELAY_SEC)
    obsolete_game_check_period_sec = attr.ib(type=int, default=OBSOLETE_GAME_CHECK_PERIOD_SEC)
    obsolete_game_check_delay_sec = attr.ib(type=int, default=OBSOLETE_GAME_CHECK_DELAY_SEC)

    def to_json(self) -> str:
        """Serialize to JSON."""
        return orjson.dumps(cattr.unstructure(self), option=orjson.OPT_INDENT_2).decode("utf-8")  # type: ignore


_CONFIG: Optional[SystemConfig] = None


def _get(
    parser: Union[Optional[ConfigParser], SectionProxy], key: str, overrides: Optional[Dict[str, Any]], defaults: Dict[str, Any]
) -> Any:
    """Get a value from a parser, overriding or setting a default as necessary."""
    override = overrides[key] if overrides and key in overrides else None
    default = defaults[key] if defaults and key in defaults else None
    return override if override else parser.get(key, default) if parser else default


# pylint: disable=too-many-locals
def _parse(
    parser: Union[Optional[ConfigParser], SectionProxy], overrides: Optional[Dict[str, Any]], defaults: Dict[str, Any]
) -> SystemConfig:
    """Create an SystemConfig based on configuration, applying defaults to values that are not available."""
    logfile_path = _get(parser, "logfile_path", overrides, defaults)
    server_host = _get(parser, "server_host", overrides, defaults)
    server_port = int(_get(parser, "server_port", overrides, defaults))
    close_timeout_sec = int(_get(parser, "close_timeout_sec", overrides, defaults))
    websocket_limit = int(_get(parser, "websocket_limit", overrides, defaults))
    total_game_limit = int(_get(parser, "total_game_limit", overrides, defaults))
    in_progress_game_limit = int(_get(parser, "in_progress_game_limit", overrides, defaults))
    registered_player_limit = int(_get(parser, "registered_player_limit", overrides, defaults))
    websocket_idle_thresh_min = int(_get(parser, "websocket_idle_thresh_min", overrides, defaults))
    websocket_inactive_thresh_min = int(_get(parser, "websocket_inactive_thresh_min", overrides, defaults))
    player_idle_thresh_min = int(_get(parser, "player_idle_thresh_min", overrides, defaults))
    player_inactive_thresh_min = int(_get(parser, "player_inactive_thresh_min", overrides, defaults))
    game_idle_thresh_min = int(_get(parser, "game_idle_thresh_min", overrides, defaults))
    game_inactive_thresh_min = int(_get(parser, "game_inactive_thresh_min", overrides, defaults))
    game_retention_thresh_min = int(_get(parser, "game_retention_thresh_min", overrides, defaults))
    idle_websocket_check_period_sec = int(_get(parser, "idle_websocket_check_period_sec", overrides, defaults))
    idle_websocket_check_delay_sec = int(_get(parser, "idle_websocket_check_delay_sec", overrides, defaults))
    idle_player_check_period_sec = int(_get(parser, "idle_player_check_period_sec", overrides, defaults))
    idle_player_check_delay_sec = int(_get(parser, "idle_player_check_delay_sec", overrides, defaults))
    idle_game_check_period_sec = int(_get(parser, "idle_game_check_period_sec", overrides, defaults))
    idle_game_check_delay_sec = int(_get(parser, "idle_game_check_delay_sec", overrides, defaults))
    obsolete_game_check_period_sec = int(_get(parser, "obsolete_game_check_period_sec", overrides, defaults))
    obsolete_game_check_delay_sec = int(_get(parser, "obsolete_game_check_delay_sec", overrides, defaults))
    return SystemConfig(
        logfile_path=logfile_path,
        server_host=server_host,
        server_port=server_port,
        close_timeout_sec=close_timeout_sec,
        websocket_limit=websocket_limit,
        total_game_limit=total_game_limit,
        in_progress_game_limit=in_progress_game_limit,
        registered_player_limit=registered_player_limit,
        websocket_idle_thresh_min=websocket_idle_thresh_min,
        websocket_inactive_thresh_min=websocket_inactive_thresh_min,
        player_idle_thresh_min=player_idle_thresh_min,
        player_inactive_thresh_min=player_inactive_thresh_min,
        game_idle_thresh_min=game_idle_thresh_min,
        game_inactive_thresh_min=game_inactive_thresh_min,
        game_retention_thresh_min=game_retention_thresh_min,
        idle_websocket_check_period_sec=idle_websocket_check_period_sec,
        idle_websocket_check_delay_sec=idle_websocket_check_delay_sec,
        idle_player_check_period_sec=idle_player_check_period_sec,
        idle_player_check_delay_sec=idle_player_check_delay_sec,
        idle_game_check_period_sec=idle_game_check_period_sec,
        idle_game_check_delay_sec=idle_game_check_delay_sec,
        obsolete_game_check_period_sec=obsolete_game_check_period_sec,
        obsolete_game_check_delay_sec=obsolete_game_check_delay_sec,
    )


def _load(config_path: str, overrides: Optional[Dict[str, Any]], defaults: Dict[str, Any]) -> SystemConfig:
    """Load configuration from disk, applying defaults for any value that is not found."""
    if not os.path.exists(config_path):
        return _parse(None, overrides, defaults)
    else:
        parser = configparser.ConfigParser()
        parser.read(config_path)
        return _parse(parser["Server"], overrides, defaults)


def load_config(config_path: Optional[str] = None, overrides: Optional[Dict[str, Any]] = None) -> None:
    """Load global configuration for later use, applying defaults for any value that is not found."""
    global _CONFIG  # pylint: disable=global-statement
    if config_path:
        # if they override the config path, the file must exist
        if not os.path.exists(config_path):
            raise ValueError("Config path does not exist: %s" % config_path)
        _CONFIG = _load(config_path, overrides, DEFAULTS)
    else:
        # however, it's ok if the default config doesn't exist
        _CONFIG = _load(DEFAULT_CONFIG_PATH, overrides, DEFAULTS)


def config() -> SystemConfig:
    """Return system configuration."""
    if not _CONFIG:
        raise ValueError("Configuration has not yet been loaded.")
    return _CONFIG



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=unsubscriptable-object

"""
Definition of the public interface for the server.

Both requests (message sent from a client to the server) and events (published 
from the server to one or more clients) can be serialized and deserialized to 
and from JSON.  However, we apply much tighter validation rules on the context
associated with requests, since the input is untrusted.  We assume that the
unit tests and the Python type validations imposed by MyPy give us everything
we need for events that are only built internally.

The file notes/API.md includes a detailed discussion of each request and event.
"""

from __future__ import annotations  # see: https://stackoverflow.com/a/33533514/2907667

from abc import ABC
from enum import Enum
from typing import Any, Dict, List, Optional, Type

import attr
import cattr
import orjson
from apologies.game import CardType, GameMode, History, Pawn, Player, PlayerColor, PlayerView, Position
from apologies.rules import Action, ActionType, Move
from attr import Attribute
from attr.validators import and_, in_
from pendulum.datetime import DateTime
from pendulum.parser import parse

from .validator import enum, length, notempty, string, stringlist

__all__ = [
    "Visibility",
    "CancelledReason",
    "PlayerType",
    "PlayerState",
    "ConnectionState",
    "ActivityState",
    "GameState",
    "FailureReason",
    "MessageType",
    "ProcessingError",
    "GamePlayer",
    "RegisteredPlayer",
    "AdvertisedGame",
    "GameStatePawn",
    "GameStatePlayer",
    "GameStateHistory",
    "GameAction",
    "GameMove",
    "RegisterPlayerContext",
    "AdvertiseGameContext",
    "JoinGameContext",
    "ExecuteMoveContext",
    "SendMessageContext",
    "RequestFailedContext",
    "RegisteredPlayersContext",
    "AvailableGamesContext",
    "PlayerRegisteredContext",
    "PlayerIdleContext",
    "PlayerInactiveContext",
    "PlayerMessageReceivedContext",
    "GameAdvertisedContext",
    "GameInvitationContext",
    "GameJoinedContext",
    "GameStartedContext",
    "GameCancelledContext",
    "GameCompletedContext",
    "GameIdleContext",
    "GameInactiveContext",
    "GamePlayerQuitContext",
    "GamePlayerChangeContext",
    "GameStateChangeContext",
    "GamePlayerTurnContext",
    "Message",
]


class Visibility(Enum):
    """Visibility for advertised games."""

    PUBLIC = "Public"
    PRIVATE = "Private"


class CancelledReason(Enum):
    """Reasons a game can be cancelled."""

    CANCELLED = "Game was cancelled by advertiser"
    NOT_VIABLE = "Game is no longer viable."
    INACTIVE = "The game was idle too long and was marked inactive"
    SHUTDOWN = "Game was cancelled due to system shutdown"


class PlayerType(Enum):
    """Types of players."""

    HUMAN = "Human"
    PROGRAMMATIC = "Programmatic"


class PlayerState(Enum):
    """A player's game state."""

    WAITING = "Waiting"
    JOINED = "Joined"
    PLAYING = "Playing"
    FINISHED = "Finished"
    QUIT = "Quit"
    DISCONNECTED = "Disconnected"


class ConnectionState(Enum):
    """A player's connection state."""

    CONNECTED = "Connected"
    DISCONNECTED = "Disconnected"


class ActivityState(Enum):
    """A player's activity state."""

    ACTIVE = "Active"
    IDLE = "Idle"
    INACTIVE = "Inactive"


class GameState(Enum):
    """A game's state."""

    ADVERTISED = "Advertised"
    PLAYING = "Playing"
    COMPLETED = "Completed"
    CANCELLED = "Cancelled"


class FailureReason(Enum):
    """Failure reasons advertised to clients."""

    INVALID_REQUEST = "Invalid request"
    DUPLICATE_USER = "Handle is already in use"
    INVALID_AUTH = "Missing or invalid authorization header"
    WEBSOCKET_LIMIT = "Connection limit reached; try again later"
    USER_LIMIT = "System user limit reached; try again later"
    GAME_LIMIT = "System game limit reached; try again later"
    INVALID_PLAYER = "Unknown or invalid player"
    INVALID_GAME = "Unknown or invalid game"
    NOT_PLAYING = "Player is not playing a game"
    NOT_ADVERTISER = "Player did not advertise this game"
    ALREADY_PLAYING = "Player is already playing a game"
    NO_MOVE_PENDING = "No move is pending for this player"
    ILLEGAL_MOVE = "The chosen move is not legal"
    ADVERTISER_MAY_NOT_QUIT = "Advertiser may not quit a game (cancel instead)"
    INTERNAL_ERROR = "Internal error"


class MessageType(Enum):
    """Enumeration of all message types, including received events and published requests."""

    # Requests sent from client to server
    REGISTER_PLAYER = "Register Player"
    REREGISTER_PLAYER = "Reregister Player"
    UNREGISTER_PLAYER = "Unregister Player"
    LIST_PLAYERS = "List Players"
    ADVERTISE_GAME = "Advertise Game"
    LIST_AVAILABLE_GAMES = "List Available"
    JOIN_GAME = "Join Game"
    QUIT_GAME = "Quit Game"
    START_GAME = "Start Game"
    CANCEL_GAME = "Cancel Game"
    EXECUTE_MOVE = "Execute Move"
    RETRIEVE_GAME_STATE = "Retrieve Game State"
    SEND_MESSAGE = "Send Message"

    # Events published from server to one or more clients
    SERVER_SHUTDOWN = "Server Shutdown"
    REQUEST_FAILED = "Request Failed"
    REGISTERED_PLAYERS = "Registered Players"
    AVAILABLE_GAMES = "Available Games"
    PLAYER_REGISTERED = "Player Registered"
    WEBSOCKET_IDLE = "Connection Idle"
    WEBSOCKET_INACTIVE = "Connection Inactive"
    PLAYER_IDLE = "Player Idle"
    PLAYER_INACTIVE = "Player Inactive"
    PLAYER_MESSAGE_RECEIVED = "Player Message Received"
    GAME_ADVERTISED = "Game Advertise"
    GAME_INVITATION = "Game Invitation"
    GAME_JOINED = "Game Joined"
    GAME_STARTED = "Game Started"
    GAME_CANCELLED = "Game Cancelled"
    GAME_COMPLETED = "Game Completed"
    GAME_IDLE = "Game Idle"
    GAME_INACTIVE = "Game Inactive"
    GAME_PLAYER_QUIT = "Game Player Quit"
    GAME_PLAYER_CHANGE = "Game Player Change"
    GAME_STATE_CHANGE = "Game State Change"
    GAME_PLAYER_TURN = "Game Player Turn"


@attr.s(frozen=True, repr=False)
class ProcessingError(RuntimeError):
    """Exception thrown when there is a general processing error."""

    reason = attr.ib(type=FailureReason)
    comment = attr.ib(type=Optional[str], default=None)
    handle = attr.ib(type=Optional[str], default=None)

    def __repr__(self) -> str:
        return self.comment if self.comment else self.reason.value  # type: ignore

    def __str__(self) -> str:
        return self.__repr__()


@attr.s(frozen=True)
class GamePlayer:
    """The public definition of a player within a game."""

    handle = attr.ib(type=str)
    player_color = attr.ib(type=Optional[PlayerColor])
    player_type = attr.ib(type=PlayerType)
    player_state = attr.ib(type=PlayerState)


@attr.s(frozen=True)
class RegisteredPlayer:
    """The public definition of a player registered with the system."""

    handle = attr.ib(type=str)
    registration_date = attr.ib(type=DateTime)
    last_active_date = attr.ib(type=DateTime)
    connection_state = attr.ib(type=ConnectionState)
    activity_state = attr.ib(type=ActivityState)
    player_state = attr.ib(type=PlayerState)
    game_id = attr.ib(type=Optional[str])


@attr.s(frozen=True)
class AdvertisedGame:
    """A game that has been advertised in the system."""

    game_id = attr.ib(type=str)
    name = attr.ib(type=str)
    mode = attr.ib(type=GameMode)
    advertiser_handle = attr.ib(type=str)
    players = attr.ib(type=int)
    available = attr.ib(type=int)
    visibility = attr.ib(type=Visibility)
    invited_handles = attr.ib(type=List[str])


@attr.s(frozen=True)
class GameStatePawn:
    """State of a pawn in a game."""

    color = attr.ib(type=PlayerColor)
    id = attr.ib(type=str)
    start = attr.ib(type=bool)
    home = attr.ib(type=bool)
    safe = attr.ib(type=Optional[int])
    square = attr.ib(type=Optional[int])

    @staticmethod
    def for_pawn(pawn: Pawn) -> GameStatePawn:
        """Create a GameStatePawn based on apologies.game.Pawn."""
        color = pawn.color
        index = "%s" % pawn.index
        start = pawn.position.start
        home = pawn.position.home
        safe = pawn.position.safe
        square = pawn.position.square
        return GameStatePawn(color, index, start, home, safe, square)

    @staticmethod
    def for_position(pawn: Pawn, position: Position) -> GameStatePawn:
        """Create a GameStatePawn based on apologies.game.Pawn and apologies.gamePosition."""
        color = pawn.color
        index = "%s" % pawn.index
        start = position.start
        home = position.home
        safe = position.safe
        square = position.square
        return GameStatePawn(color, index, start, home, safe, square)


@attr.s(frozen=True)
class GameStatePlayer:
    """Player in a game, when describing the state of the board."""

    color = attr.ib(type=PlayerColor)
    turns = attr.ib(type=int)
    hand = attr.ib(type=List[CardType])
    pawns = attr.ib(type=List[GameStatePawn])

    @staticmethod
    def for_player(player: Player) -> GameStatePlayer:
        """Create a GameStatePlayer based on apologies.game.Player."""
        color = player.color
        turns = player.turns
        hand = [card.cardtype for card in player.hand]
        pawns = [GameStatePawn.for_pawn(pawn) for pawn in player.pawns]
        return GameStatePlayer(color, turns, hand, pawns)


@attr.s
class GameStateHistory:
    """History for a game."""

    action = attr.ib(type=str)
    color = attr.ib(type=Optional[PlayerColor])
    timestamp = attr.ib(type=DateTime)

    @staticmethod
    def for_history(history: History) -> GameStateHistory:
        return GameStateHistory(action=history.action, color=history.color, timestamp=history.timestamp)


@attr.s(frozen=True)
class GameAction:
    """An action applied to a pawn in a game."""

    start = attr.ib(type=GameStatePawn)
    end = attr.ib(type=GameStatePawn)

    @staticmethod
    def for_action(action: Action) -> GameAction:
        """Create a GamePlayerAction based on apologies.rules.Action."""
        if action.actiontype == ActionType.MOVE_TO_START:
            # We normalize a MOVE_TO_START action to just a position change, to simplify what the client sees
            start = GameStatePawn.for_pawn(action.pawn)
            end = GameStatePawn.for_position(action.pawn, Position().move_to_start())
            return GameAction(start, end)
        else:  # action.actiontype == ActionType.MOVE_TO_POSITION
            start = GameStatePawn.for_pawn(action.pawn)
            end = GameStatePawn.for_position(action.pawn, action.position)
            return GameAction(start, end)


@attr.s(frozen=True)
class GameMove:
    """A move that may be executed as a result of a player's turn."""

    move_id = attr.ib(type=str)
    card = attr.ib(type=CardType)
    actions = attr.ib(type=List[GameAction])
    side_effects = attr.ib(type=List[GameAction])

    @staticmethod
    def for_move(move: Move) -> GameMove:
        """Create a GameMove based on apologies.rules.Move."""
        move_id = move.id
        card = move.card.cardtype
        actions = [GameAction.for_action(action) for action in move.actions]
        side_effects = [GameAction.for_action(side_effect) for side_effect in move.side_effects]
        return GameMove(move_id, card, actions, side_effects)


class Context(ABC):
    """Abstract message context."""


MAX_HANDLE = 25
"""Maximum length of a player handle."""


@attr.s(frozen=True)
class RegisterPlayerContext(Context):
    """Context for a REGISTER_PLAYER request."""

    handle = attr.ib(type=str, validator=and_(string, length(MAX_HANDLE)))


@attr.s(frozen=True)
class AdvertiseGameContext(Context):
    """Context for an ADVERTISE_GAME request."""

    name = attr.ib(type=str, validator=string)
    mode = attr.ib(type=GameMode, validator=enum(GameMode))
    players = attr.ib(type=int, validator=in_([2, 3, 4]))
    visibility = attr.ib(type=Visibility, validator=enum(Visibility))
    invited_handles = attr.ib(type=List[str], validator=stringlist)


@attr.s(frozen=True)
class JoinGameContext(Context):
    """Context for a JOIN_GAME request."""

    game_id = attr.ib(type=str, validator=string)


@attr.s(frozen=True)
class ExecuteMoveContext(Context):
    """Context for an EXECUTE_MOVE request."""

    move_id = attr.ib(type=str, validator=string)


@attr.s(frozen=True)
class SendMessageContext(Context):
    """Context for an SEND_MESSAGE request."""

    message = attr.ib(type=str, validator=string)
    recipient_handles = attr.ib(type=List[str], validator=and_(stringlist, notempty))


@attr.s(frozen=True)
class RequestFailedContext(Context):
    """Context for a REQUEST_FAILED event."""

    reason = attr.ib(type=FailureReason)
    comment = attr.ib(type=Optional[str])
    handle = attr.ib(type=Optional[str], default=None)


@attr.s(frozen=True)
class RegisteredPlayersContext(Context):
    """Context for a REGISTERED_PLAYERS event."""

    players = attr.ib(type=List[RegisteredPlayer])


@attr.s(frozen=True)
class AvailableGamesContext(Context):
    """Context for an AVAILABLE_GAMES event."""

    games = attr.ib(type=List[AdvertisedGame])


@attr.s(frozen=True)
class PlayerRegisteredContext(Context):
    """Context for a PLAYER_REGISTERED event."""

    handle = attr.ib(type=str)


@attr.s(frozen=True)
class PlayerIdleContext(Context):
    """Context for a PLAYER_IDLE event."""

    handle = attr.ib(type=str)


@attr.s(frozen=True)
class PlayerInactiveContext(Context):
    """Context for a PLAYER_INACTIVE event."""

    handle = attr.ib(type=str)


@attr.s(frozen=True)
class PlayerMessageReceivedContext(Context):
    """Context for a PLAYER_MESSAGE_RECEIVED event."""

    sender_handle = attr.ib(type=str)
    recipient_handles = attr.ib(type=List[str])
    message = attr.ib(type=str)


@attr.s(frozen=True)
class GameAdvertisedContext(Context):
    """Context for a GAME_ADVERTISED event."""

    game = attr.ib(type=AdvertisedGame)


@attr.s(frozen=True)
class GameInvitationContext(Context):
    """Context for a GAME_INVITATION event."""

    game = attr.ib(type=AdvertisedGame)


@attr.s(frozen=True)
class GameJoinedContext(Context):
    """Context for a GAME_JOINED event."""

    game_id = attr.ib(type=str)
    # TODO: we need the handle of the player that joined, in case multiple players share the same websocket


@attr.s(frozen=True)
class GameStartedContext(Context):
    """Context for a GAME_STARTED event."""

    game_id = attr.ib(type=str)


@attr.s(frozen=True)
class GameCancelledContext(Context):
    """Context for a GAME_CANCELLED event."""

    game_id = attr.ib(type=str)
    reason = attr.ib(type=CancelledReason)
    comment = attr.ib(type=Optional[str])


@attr.s(frozen=True)
class GameCompletedContext(Context):
    """Context for a GAME_COMPLETED event."""

    game_id = attr.ib(type=str)
    comment = attr.ib(type=Optional[str])


@attr.s(frozen=True)
class GameIdleContext(Context):
    """Context for a GAME_IDLE event."""

    game_id = attr.ib(type=str)


@attr.s(frozen=True)
class GameInactiveContext(Context):
    """Context for a GAME_INACTIVE event."""

    game_id = attr.ib(type=str)


@attr.s(frozen=True)
class GamePlayerQuitContext(Context):
    """Context for a GAME_PLAYER_LEFT event."""

    handle = attr.ib(type=str)
    game_id = attr.ib(type=str)


@attr.s(frozen=True)
class GamePlayerChangeContext(Context):
    """Context for a GAME_PLAYER_CHANGE event."""

    game_id = attr.ib(type=str)
    comment = attr.ib(type=Optional[str])
    players = attr.ib(type=List[GamePlayer])


@attr.s(frozen=True)
class GameStateChangeContext(Context):
    """Context for a GAME_STATE_CHANGE event."""

    game_id = attr.ib(type=str)
    recent_history = attr.ib(type=List[GameStateHistory])
    player = attr.ib(type=GameStatePlayer)
    opponents = attr.ib(type=List[GameStatePlayer])

    @staticmethod
    def for_context(game_id: str, view: PlayerView, history: List[History]) -> GameStateChangeContext:
        """Create a GameStateChangeContext based on apologies.game.PlayerView."""
        player = GameStatePlayer.for_player(view.player)
        recent_history = [GameStateHistory.for_history(entry) for entry in history]
        opponents = [GameStatePlayer.for_player(opponent) for opponent in view.opponents.values()]
        return GameStateChangeContext(game_id, recent_history, player, opponents)


@attr.s(frozen=True)
class GamePlayerTurnContext(Context):
    """Context for a GAME_PLAYER_TURN event."""

    handle = attr.ib(type=str)
    game_id = attr.ib(type=str)
    drawn_card = attr.ib(type=Optional[CardType])
    moves = attr.ib(type=Dict[str, GameMove])

    @staticmethod
    def for_moves(handle: str, game_id: str, moves: List[Move]) -> GamePlayerTurnContext:
        """Create a GamePlayerTurnContext based on a sequence of apologies.rules.Move."""
        cards = {move.card.cardtype for move in moves}
        drawn_card = None if len(cards) > 1 else next(iter(cards))  # if there's only one card, it's the one they drew from the deck
        converted = {move.id: GameMove.for_move(move) for move in moves}
        return GamePlayerTurnContext(handle, game_id, drawn_card, converted)


# Map from MessageType to whether player id field is allowed/required
_PLAYER_ID: Dict[MessageType, bool] = {
    MessageType.REGISTER_PLAYER: False,
    MessageType.REREGISTER_PLAYER: True,
    MessageType.UNREGISTER_PLAYER: True,
    MessageType.LIST_PLAYERS: True,
    MessageType.ADVERTISE_GAME: True,
    MessageType.LIST_AVAILABLE_GAMES: True,
    MessageType.JOIN_GAME: True,
    MessageType.QUIT_GAME: True,
    MessageType.START_GAME: True,
    MessageType.CANCEL_GAME: True,
    MessageType.EXECUTE_MOVE: True,
    MessageType.RETRIEVE_GAME_STATE: True,
    MessageType.SEND_MESSAGE: True,
    MessageType.SERVER_SHUTDOWN: False,
    MessageType.REQUEST_FAILED: False,
    MessageType.WEBSOCKET_IDLE: False,
    MessageType.WEBSOCKET_INACTIVE: False,
    MessageType.REGISTERED_PLAYERS: False,
    MessageType.AVAILABLE_GAMES: False,
    MessageType.PLAYER_REGISTERED: True,
    MessageType.PLAYER_IDLE: False,
    MessageType.PLAYER_INACTIVE: False,
    MessageType.PLAYER_MESSAGE_RECEIVED: False,
    MessageType.GAME_ADVERTISED: False,
    MessageType.GAME_INVITATION: False,
    MessageType.GAME_JOINED: False,
    MessageType.GAME_STARTED: False,
    MessageType.GAME_CANCELLED: False,
    MessageType.GAME_COMPLETED: False,
    MessageType.GAME_IDLE: False,
    MessageType.GAME_INACTIVE: False,
    MessageType.GAME_PLAYER_QUIT: False,
    MessageType.GAME_PLAYER_CHANGE: False,
    MessageType.GAME_STATE_CHANGE: False,
    MessageType.GAME_PLAYER_TURN: False,
}

# Map from MessageType to context
_CONTEXT: Dict[MessageType, Optional[Type[Context]]] = {
    MessageType.REGISTER_PLAYER: RegisterPlayerContext,
    MessageType.REREGISTER_PLAYER: None,
    MessageType.UNREGISTER_PLAYER: None,
    MessageType.LIST_PLAYERS: None,
    MessageType.ADVERTISE_GAME: AdvertiseGameContext,
    MessageType.LIST_AVAILABLE_GAMES: None,
    MessageType.JOIN_GAME: JoinGameContext,
    MessageType.QUIT_GAME: None,
    MessageType.START_GAME: None,
    MessageType.CANCEL_GAME: None,
    MessageType.EXECUTE_MOVE: ExecuteMoveContext,
    MessageType.RETRIEVE_GAME_STATE: None,
    MessageType.SEND_MESSAGE: SendMessageContext,
    MessageType.SERVER_SHUTDOWN: None,
    MessageType.REQUEST_FAILED: RequestFailedContext,
    MessageType.WEBSOCKET_IDLE: None,
    MessageType.WEBSOCKET_INACTIVE: None,
    MessageType.REGISTERED_PLAYERS: RegisteredPlayersContext,
    MessageType.AVAILABLE_GAMES: AvailableGamesContext,
    MessageType.PLAYER_REGISTERED: PlayerRegisteredContext,
    MessageType.PLAYER_IDLE: PlayerIdleContext,
    MessageType.PLAYER_INACTIVE: PlayerInactiveContext,
    MessageType.PLAYER_MESSAGE_RECEIVED: PlayerMessageReceivedContext,
    MessageType.GAME_ADVERTISED: GameAdvertisedContext,
    MessageType.GAME_INVITATION: GameInvitationContext,
    MessageType.GAME_JOINED: GameJoinedContext,
    MessageType.GAME_STARTED: GameStartedContext,
    MessageType.GAME_CANCELLED: GameCancelledContext,
    MessageType.GAME_COMPLETED: GameCompletedContext,
    MessageType.GAME_IDLE: GameIdleContext,
    MessageType.GAME_INACTIVE: GameInactiveContext,
    MessageType.GAME_PLAYER_QUIT: GamePlayerQuitContext,
    MessageType.GAME_PLAYER_CHANGE: GamePlayerChangeContext,
    MessageType.GAME_STATE_CHANGE: GameStateChangeContext,
    MessageType.GAME_PLAYER_TURN: GamePlayerTurnContext,
}

# List of all enumerations that are part of the public interface
_ENUMS = [
    Visibility,
    FailureReason,
    CancelledReason,
    PlayerType,
    PlayerState,
    ConnectionState,
    ActivityState,
    MessageType,
    GameMode,
    PlayerColor,
    CardType,
]

_DATE_FORMAT = "YYYY-MM-DDTHH:mm:ss,SSSZ"  # gives us something like "2020-04-27T09:02:14,334+00:00"


class _CattrConverter(cattr.Converter):  # type: ignore
    """
    Cattr converter for requests and events, to standardize conversion of dates and enumerations.
    """

    def __init__(self) -> None:
        super().__init__()
        self.register_unstructure_hook(DateTime, lambda value: value.format(_DATE_FORMAT) if value else None)
        self.register_structure_hook(DateTime, lambda value, _: parse(value) if value else None)
        for element in _ENUMS:
            self.register_unstructure_hook(element, lambda value: value.name if value else None)
            self.register_structure_hook(element, lambda value, _, e=element: e[value] if value else None)


# Cattr converter used to serialize and deserialize requests and responses
_CONVERTER = _CattrConverter()


# noinspection PyTypeChecker
@attr.s(frozen=True)
class Message:
    """A message that is part of the public interface, either a client request or a published event."""

    message = attr.ib(type=MessageType)
    player_id = attr.ib(type=Optional[str], default=None, repr=False)  # this is a secret, so we don't want it printed or logged
    context = attr.ib(type=Any, default=None)

    @message.validator
    def _validate_message(self, attribute: Attribute[MessageType], value: MessageType) -> None:
        if value is None or not isinstance(value, MessageType):
            raise ValueError("'%s' must be a MessageType" % attribute.name)

    @player_id.validator
    def _validate_player_id(self, _attribute: Attribute[str], value: str) -> None:
        if _PLAYER_ID[self.message]:
            if value is None:
                raise ValueError("Message type %s requires a player id" % self.message.name)
        else:
            if value is not None:
                raise ValueError("Message type %s does not allow a player id" % self.message.name)

    @context.validator
    def _validate_context(self, _attribute: Attribute[Context], value: Context) -> None:
        if _CONTEXT[self.message] is not None:
            if value is None:
                raise ValueError("Message type %s requires a context" % self.message.name)
            elif not isinstance(value, _CONTEXT[self.message]):  # type: ignore
                raise ValueError("Message type %s does not support this context" % self.message.name)
        else:
            if value is not None:
                raise ValueError("Message type %s does not allow a context" % self.message.name)

    def to_json(self) -> str:
        """Convert the request to JSON."""
        d = _CONVERTER.unstructure(self)  # pylint: disable=invalid-name
        if d["player_id"] is None:
            del d["player_id"]
        if d["context"] is None:
            del d["context"]
        return orjson.dumps(d, option=orjson.OPT_INDENT_2).decode("utf-8")  # type: ignore

    @staticmethod
    def for_json(data: str) -> Message:
        """Create a request based on JSON data."""
        d = orjson.loads(data)  # pylint: disable=invalid-name
        if "message" not in d or d["message"] is None:
            raise ValueError("Message type is required")
        try:
            message = MessageType[d["message"]]
        except KeyError:
            raise ValueError("Unknown message type: %s" % d["message"])
        if _PLAYER_ID[message]:
            if "player_id" not in d or d["player_id"] is None:
                raise ValueError("Message type %s requires a player id" % message.name)
            player_id = d["player_id"]
        else:
            if "player_id" in d and d["player_id"] is not None:
                raise ValueError("Message type %s does not allow a player id" % message.name)
            player_id = None
        if _CONTEXT[message] is None:
            if "context" in d and d["context"] is not None:
                raise ValueError("Message type %s does not allow a context" % message.name)
            context = None
        else:
            if "context" not in d or d["context"] is None:
                raise ValueError("Message type %s requires a context" % message.name)
            try:
                context = _CONVERTER.structure(d["context"], _CONTEXT[message])
            except KeyError as e:
                raise ValueError("Invalid value %s" % str(e))
            except TypeError as e:
                raise ValueError("Message type %s does not support this context" % message.name, e)
        return Message(message, player_id, context)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=wildcard-import

"""Coroutines to handle scheduled tasks, executed on a periodic basis."""

import logging
from typing import Any, Callable, Coroutine, List

from periodic import Periodic

from .config import config
from .event import *
from .event import EventHandler
from .interface import ConnectionState, GameState
from .manager import manager

log = logging.getLogger("apologies.scheduled")


async def _execute_idle_websocket_check() -> None:
    """Execute the Idle Websocket Check task."""
    with EventHandler(manager()) as handler:
        async with handler.manager.lock:
            handler.handle_idle_websocket_check_task()
        await handler.execute_tasks()


async def _execute_idle_player_check() -> None:
    """Execute the Idle Player Check task."""
    with EventHandler(manager()) as handler:
        async with handler.manager.lock:
            handler.handle_idle_player_check_task()
        await handler.execute_tasks()


async def _execute_idle_game_check() -> None:
    """Execute the Idle Game Check task."""
    with EventHandler(manager()) as handler:
        async with handler.manager.lock:
            handler.handle_idle_game_check_task()
        await handler.execute_tasks()


async def _execute_obsolete_game_check() -> None:
    """Execute the Obsolete Game Check task."""
    with EventHandler(manager()) as handler:
        async with handler.manager.lock:
            handler.handle_obsolete_game_check_task()
        await handler.execute_tasks()


async def _schedule_idle_websocket_check() -> None:
    """Schedule the Idle Websocket Check task to run periodically, with a delay before starting."""
    period = config().idle_websocket_check_period_sec
    delay = config().idle_websocket_check_delay_sec
    p = Periodic(period, _execute_idle_websocket_check)
    await p.start(delay=delay)
    log.debug("Completed scheduling idle websocket check with period %d and delay %d", period, delay)


async def _schedule_idle_player_check() -> None:
    """Schedule the Idle Player Check task to run periodically, with a delay before starting."""
    period = config().idle_player_check_period_sec
    delay = config().idle_player_check_delay_sec
    p = Periodic(period, _execute_idle_player_check)
    await p.start(delay=delay)
    log.debug("Completed scheduling idle player check with period %d and delay %d", period, delay)


async def _schedule_idle_game_check() -> None:
    """Schedule the Idle Game Check task to run periodically, with a delay before starting."""
    period = config().idle_game_check_period_sec
    delay = config().idle_game_check_delay_sec
    p = Periodic(period, _execute_idle_game_check)
    await p.start(delay=delay)
    log.debug("Completed scheduling idle game check with period %d and delay %d", period, delay)


async def _schedule_obsolete_game_check() -> None:
    """Schedule the Obsolete Check task to run periodically, with a delay before starting."""
    period = config().obsolete_game_check_period_sec
    delay = config().obsolete_game_check_delay_sec
    p = Periodic(period, _execute_obsolete_game_check)
    await p.start(delay=delay)
    log.debug("Completed scheduling obsolete game check with period %d and delay %d", period, delay)


def scheduled_tasks() -> List[Callable[[], Coroutine[Any, Any, None]]]:
    """Get a list of tasks that need to be scheduled."""
    return [_schedule_idle_websocket_check, _schedule_idle_player_check, _schedule_idle_game_check, _schedule_obsolete_game_check]



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=unsubscriptable-object

"""
Shared utilities.
"""

import asyncio
import logging
import re
import sys
import time
from asyncio import TimeoutError  # pylint: disable=redefined-builtin
from pathlib import Path
from typing import Optional, Union, cast

from websockets import WebSocketCommonProtocol
from websockets.typing import Data

from .interface import Message, MessageType, ProcessingError, RequestFailedContext

log = logging.getLogger("apologies.util")


def homedir() -> str:
    """Get the current user's home directory."""
    return str(Path.home())


def mask(data: Optional[Union[str, bytes]]) -> str:
    """Mask the player id in JSON data, since it's a secret we don't want logged."""
    decoded = "" if not data else data.decode("utf-8") if isinstance(data, bytes) else data
    return re.sub(r'"player_id" *: *"[^"]+"', r'"player_id": "<masked>"', decoded)


def extract(data: Union[str, Message, Data]) -> Message:
    message = Message.for_json(str(data))
    if message.message == MessageType.REQUEST_FAILED:
        context = cast(RequestFailedContext, message.context)
        raise ProcessingError(reason=context.reason, comment=context.comment, handle=context.handle)
    return message


async def close(websocket: WebSocketCommonProtocol) -> None:
    """Close a websocket."""
    log.debug("Closing websocket: %s", id(websocket))
    await websocket.close()


async def send(websocket: WebSocketCommonProtocol, message: Union[str, Message]) -> None:
    """Send a response to a websocket."""
    if message:
        data = message.to_json() if isinstance(message, Message) else message
        log.debug("Sending message to websocket: %s\n%s", id(websocket), mask(data))
        await websocket.send(data)


async def receive(websocket: WebSocketCommonProtocol, timeout_sec: Optional[int] = None) -> Optional[Message]:
    try:
        data = await websocket.recv() if not timeout_sec else await asyncio.wait_for(websocket.recv(), timeout=timeout_sec)
        log.debug("Received raw data for websocket %s:\n%s", id(websocket), mask(data))
        return extract(data)
    except TimeoutError:
        log.debug("Timed out waiting for raw data for websocket %s", id(websocket))
        return None


def setup_logging(quiet: bool, verbose: bool, debug: bool, logfile_path: Optional[str] = None) -> None:
    """Set up Python logging."""
    logger = logging.getLogger("apologies")
    logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(logfile_path) if logfile_path else logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(fmt="%(asctime)sZ --> [%(levelname)-7s] %(message)s")
    formatter.converter = time.gmtime  # type: ignore
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)
    if quiet:
        handler.setLevel(logging.ERROR)
    if verbose or debug:
        handler.setLevel(logging.DEBUG)
    if debug:
        wslogger = logging.getLogger("websockets")
        wslogger.setLevel(logging.INFO)
        wslogger.addHandler(handler)
    logger.addHandler(handler)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=wildcard-import

"""
Event handlers.

All event handlers operate in terms of passed-in state, encapsulated in the StateManager
object.  The caller must ensure that 1) the state manager is locked before calling any
handler methods other than execute() and 2) that the execute() method is called once 
at the end of processing but outside of the manager lock.  The goal is to ensure that 
state-related operations are fast and non-blocking, and happen behind a clear transaction 
boundary.  Any tasks that might block (such as network requests) should be added to the 
task queue, to be executed by the execute() method once state updates have been completed.  
"""

from __future__ import annotations  # see: https://stackoverflow.com/a/33533514/2907667

import asyncio
import logging
from typing import List, Optional, Set, Tuple, cast

import attr
import pendulum
from apologies.rules import Move, Rules
from apologies.source import RewardV1InputSource
from ordered_set import OrderedSet  # this makes expected results easier to articulate in test code
from websockets import WebSocketServerProtocol

from .config import config
from .interface import *
from .manager import StateManager, TrackedGame, TrackedPlayer, TrackedWebsocket
from .util import close, send

log = logging.getLogger("apologies.event")


@attr.s(frozen=True)
class RequestContext:
    """Context provided to a request handler method."""

    message = attr.ib(type=Message)
    websocket = attr.ib(type=WebSocketServerProtocol)
    player = attr.ib(type=TrackedPlayer)
    game = attr.ib(type=Optional[TrackedGame], default=None)


# noinspection PyMethodMayBeStatic
@attr.s
class TaskQueue:

    """A queue of asynchronous tasks to be executed."""

    messages = attr.ib(type=List[Tuple[str, WebSocketServerProtocol]])
    disconnects = attr.ib(type=Set[WebSocketServerProtocol])

    @messages.default
    def _messages_default(self) -> List[Tuple[str, WebSocketServerProtocol]]:
        return []

    @disconnects.default
    def _disconnects_default(self) -> OrderedSet[WebSocketServerProtocol]:
        return OrderedSet()

    def is_empty(self) -> bool:
        return len(self.messages) == 0 and len(self.disconnects) == 0

    def clear(self) -> None:
        del self.messages[:]
        self.disconnects.clear()

    def message(
        self,
        message: Message,
        websockets: Optional[List[WebSocketServerProtocol]] = None,
        players: Optional[List[TrackedPlayer]] = None,
    ) -> None:
        """Enqueue a task to send a message to one or more destination websockets."""
        destinations = OrderedSet(websockets) if websockets else OrderedSet()
        destinations.update([player.websocket for player in players if player.websocket] if players else [])
        self.messages.extend([(message.to_json(), destination) for destination in destinations])

    def disconnect(self, websocket: Optional[WebSocketServerProtocol]) -> None:
        """Enqueue a task to disconnect a websocket."""
        if websocket:
            self.disconnects.add(websocket)

    async def execute(self) -> None:
        """Execute all tasks in the queue, sending messages first and then disconnecting websockets."""
        tasks = [send(websocket, message) for message, websocket in self.messages]
        if tasks:
            await asyncio.wait(tasks)
        tasks = [close(websocket) for websocket in self.disconnects]
        if tasks:
            await asyncio.wait(tasks)  # TODO: not entirely sure how we handle errors that happen here


# pylint: disable=too-many-public-methods
@attr.s
class EventHandler:

    manager = attr.ib(type=StateManager)
    queue = attr.ib(type=TaskQueue)

    @queue.default
    def _default_queue(self) -> TaskQueue:
        return TaskQueue()

    def __enter__(self) -> EventHandler:
        self.queue.clear()
        return self

    def __exit__(self, _type, _value, _tb) -> None:  # type: ignore
        self.queue.clear()

    async def execute_tasks(self) -> None:
        """Send all enqueued tasks."""
        # This should not be invoked from within the manager lock!  We want code within the lock to run fast, without blocking.
        await self.queue.execute()

    # noinspection PyTypeChecker
    def handle_idle_websocket_check_task(self) -> Tuple[int, int]:
        """Execute the Idle Websocket Check task, returning tuple of (idle, inactive)."""
        log.info("Scheduled - IDLE_WEBSOCKET_CHECK")
        idle = 0
        inactive = 0
        now = pendulum.now()
        idle_thresh_min = config().websocket_idle_thresh_min
        inactive_thresh_min = config().websocket_inactive_thresh_min
        for (websocket, last_active_date, players) in self.manager.lookup_websocket_activity():
            if players < 1:  # by definition, a websocket is not idle until or unless it has no registered players
                since_active = now.diff(last_active_date).in_minutes()
                if since_active >= inactive_thresh_min:
                    inactive += 1
                    self.handle_websocket_inactive_event(websocket)
                elif since_active >= idle_thresh_min:
                    idle += 1
                    self.handle_websocket_idle_event(websocket)
        log.debug("Idle websocket check completed, found %d idle and %d inactive websockets", idle, inactive)
        return idle, inactive

    # noinspection PyTypeChecker
    def handle_idle_player_check_task(self) -> Tuple[int, int]:
        """Execute the Idle Player Check task, returning tuple of (idle, inactive)."""
        log.info("Scheduled - IDLE_PLAYER_CHECK")
        idle = 0
        inactive = 0
        now = pendulum.now()
        idle_thresh_min = config().player_idle_thresh_min
        inactive_thresh_min = config().player_inactive_thresh_min
        for (player, last_active_date, connection_state) in self.manager.lookup_player_activity():
            disconnected = connection_state == ConnectionState.DISCONNECTED
            since_active = now.diff(last_active_date).in_minutes()
            if since_active >= inactive_thresh_min:
                inactive += 1
                self.handle_player_inactive_event(player)
            elif since_active >= idle_thresh_min:
                if disconnected:
                    inactive += 1
                    self.handle_player_inactive_event(player)
                else:
                    idle += 1
                    self.handle_player_idle_event(player)
        log.debug("Idle player check completed, found %d idle and %d inactive players", idle, inactive)
        return idle, inactive

    # noinspection PyTypeChecker
    def handle_idle_game_check_task(self) -> Tuple[int, int]:
        """Execute the Idle Game Check task, returning tuple of (idle, inactive)."""
        log.info("Scheduled - IDLE_GAME_CHECK")
        idle = 0
        inactive = 0
        now = pendulum.now()
        idle_thresh_min = config().game_idle_thresh_min
        inactive_thresh_min = config().game_inactive_thresh_min
        for (game, last_active_date) in self.manager.lookup_game_activity():
            since_active = now.diff(last_active_date).in_minutes()
            if since_active >= inactive_thresh_min:
                inactive += 1
                self.handle_game_inactive_event(game)
            elif since_active >= idle_thresh_min:
                idle += 1
                self.handle_game_idle_event(game)
        log.debug("Idle game check completed, found %d idle and %d inactive games", idle, inactive)
        return idle, inactive

    # noinspection PyTypeChecker
    def handle_obsolete_game_check_task(self) -> int:
        """Execute the Obsolete Game Check task returning obsolete games."""
        log.info("Scheduled - OBSOLETE_GAME_CHECK")
        obsolete = 0
        now = pendulum.now()
        retention_thresh_min = config().game_retention_thresh_min
        for (game, completed_date) in self.manager.lookup_game_completion():
            since_completed = now.diff(completed_date).in_minutes()
            if since_completed >= retention_thresh_min:
                obsolete += 1
                self.handle_game_obsolete_event(game)
        log.debug("Obsolete game check completed, found %d obsolete games", obsolete)
        return obsolete

    def handle_register_player_request(self, message: Message, websocket: WebSocketServerProtocol) -> None:
        """Handle the Register Player request."""
        context = cast(RegisterPlayerContext, message.context)
        log.info("Request - REGISTER PLAYER - %s on %s", context.handle, id(websocket))
        if self.manager.get_registered_player_count() >= config().registered_player_limit:
            raise ProcessingError(FailureReason.USER_LIMIT, handle=context.handle)
        self.handle_player_registered_event(websocket, context.handle)

    def handle_reregister_player_request(self, request: RequestContext) -> None:
        """Handle the Reregister Player request."""
        log.info("Request - REREGISTER PLAYER - %s", request.player.handle)
        self.handle_player_reregistered_event(request.player, request.websocket)

    def handle_unregister_player_request(self, request: RequestContext) -> None:
        """Handle the Unregister Player request."""
        log.info("Request - UNREGISTER PLAYER - %s", request.player.handle)
        self.handle_player_unregistered_event(request.player, request.game)

    def handle_list_players_request(self, request: RequestContext) -> None:
        """Handle the List Players request."""
        log.info("Request - LIST PLAYERS - %s", request.player.handle)
        self.handle_registered_players_event(request.player)

    def handle_advertise_game_request(self, request: RequestContext) -> None:
        """Handle the Advertise Game request."""
        log.info("Request - ADVERTISE GAME - %s", request.player.handle)
        if request.game:
            raise ProcessingError(FailureReason.ALREADY_PLAYING, handle=request.player.handle)
        if self.manager.get_total_game_count() >= config().total_game_limit:
            raise ProcessingError(FailureReason.GAME_LIMIT, handle=request.player.handle)
        context = cast(AdvertiseGameContext, request.message.context)
        self.handle_game_advertised_event(request.player, context)

    def handle_list_available_games_request(self, request: RequestContext) -> None:
        """Handle the List Available Games request."""
        log.info("Request - LIST AVAILABLE_GAMES - %s", request.player.handle)
        self.handle_available_games_event(request.player)

    def handle_join_game_request(self, request: RequestContext) -> None:
        """Handle the Join Game request."""
        context = cast(JoinGameContext, request.message.context)
        log.info("Request - JOIN GAME - %s for %s", request.player.handle, context.game_id)
        if request.game:
            raise ProcessingError(FailureReason.ALREADY_PLAYING, handle=request.player.handle)
        self.handle_game_joined_event(request.player, game_id=context.game_id)

    def handle_quit_game_request(self, request: RequestContext) -> None:
        """Handle the Quit Game request."""
        log.info("Request - QUIT GAME - %s quit %s", request.player.handle, request.player.game_id)
        if not request.game:
            raise ProcessingError(FailureReason.NOT_PLAYING, handle=request.player.handle)
        if not request.game.is_in_progress():
            raise ProcessingError(FailureReason.INVALID_GAME, comment="Game is not in progress", handle=request.player.handle)
        if request.player.handle == request.game.advertiser_handle:
            raise ProcessingError(FailureReason.ADVERTISER_MAY_NOT_QUIT, handle=request.player.handle)
        self.handle_game_player_quit_event(request.player, request.game)

    def handle_start_game_request(self, request: RequestContext) -> None:
        """Handle the Start Game request."""
        log.info("Request - START GAME - %s started %s", request.player.handle, request.player.game_id)
        if not request.game:
            raise ProcessingError(FailureReason.NOT_PLAYING, handle=request.player.handle)
        if request.game.is_playing():
            raise ProcessingError(FailureReason.INVALID_GAME, comment="Game is already being played", handle=request.player.handle)
        if request.game.advertiser_handle != request.player.handle:
            raise ProcessingError(FailureReason.NOT_ADVERTISER, handle=request.player.handle)
        if self.manager.get_in_progress_game_count() >= config().in_progress_game_limit:
            raise ProcessingError(FailureReason.GAME_LIMIT, handle=request.player.handle)
        self.handle_game_started_event(request.game)

    def handle_cancel_game_request(self, request: RequestContext) -> None:
        """Handle the Cancel Game request."""
        log.info("Request - CANCEL GAME - %s cancelled %s", request.player.handle, request.player.game_id)
        if not request.game:
            raise ProcessingError(FailureReason.NOT_PLAYING, handle=request.player.handle)
        if not request.game.is_in_progress():
            raise ProcessingError(FailureReason.INVALID_GAME, comment="Game is not in progress", handle=request.player.handle)
        if request.game.advertiser_handle != request.player.handle:
            raise ProcessingError(FailureReason.NOT_ADVERTISER, handle=request.player.handle)
        self.handle_game_cancelled_event(request.game, CancelledReason.CANCELLED)

    def handle_execute_move_request(self, request: RequestContext) -> None:
        """Handle the Execute Move request."""
        context = cast(ExecuteMoveContext, request.message.context)
        log.info("Request - EXECUTE MOVE - %s for %s, move %s", request.player.handle, request.player.game_id, context.move_id)
        if not request.game:
            raise ProcessingError(FailureReason.NOT_PLAYING, handle=request.player.handle)
        if not request.game.is_playing():
            raise ProcessingError(FailureReason.INVALID_GAME, comment="Game is not being played", handle=request.player.handle)
        if not request.game.is_move_pending(request.player.handle):
            raise ProcessingError(FailureReason.NO_MOVE_PENDING, handle=request.player.handle)
        if not request.game.is_legal_move(request.player.handle, context.move_id):
            raise ProcessingError(FailureReason.ILLEGAL_MOVE, handle=request.player.handle)
        self.handle_game_player_move_event(request.player, request.game, context.move_id)

    def handle_retrieve_game_state_request(self, request: RequestContext) -> None:
        """Handle the Retrieve Game State request."""
        log.info("Request - RETRIEVE GAME - %s for %s", request.player.handle, request.player.game_id)
        if not request.game:
            raise ProcessingError(FailureReason.NOT_PLAYING, handle=request.player.handle)
        if not request.game.is_playing():
            raise ProcessingError(FailureReason.INVALID_GAME, comment="Game is not being played", handle=request.player.handle)
        self.handle_game_state_change_event(request.game, request.player)

    def handle_send_message_request(self, request: RequestContext) -> None:
        """Handle the Send Message request."""
        context = cast(SendMessageContext, request.message.context)
        log.info(
            "Request - SEND MESSAGE - %s sending %d bytes to %d recipients",
            request.player.handle,
            len(context.message),
            len(context.recipient_handles),
        )
        self.handle_player_message_received_event(request.player.handle, context.recipient_handles, context.message)

    def handle_server_shutdown_event(self) -> None:
        """Handle the Server Shutdown event."""
        log.info("Event - SERVER SHUTDOWN")
        websockets = self.manager.lookup_all_websockets()
        message = Message(MessageType.SERVER_SHUTDOWN)
        self.queue.message(message, websockets=websockets)
        for game in self.manager.lookup_in_progress_games():
            self.handle_game_cancelled_event(game, CancelledReason.SHUTDOWN, notify=False)

    def handle_websocket_connected_event(self, websocket: WebSocketServerProtocol) -> None:
        """Handle the Websocket Connected event."""
        log.info("Event - WEBSOCKET CONNECTED - %s", id(websocket))
        if self.manager.get_websocket_count() >= config().websocket_limit:
            raise ProcessingError(FailureReason.WEBSOCKET_LIMIT)
        self.manager.track_websocket(websocket)

    def handle_websocket_disconnected_event(self, websocket: WebSocketServerProtocol) -> None:
        """Handle the Websocket Disconnected event."""
        log.info("Event - WEBSOCKET DISCONNECTED - %s", id(websocket))
        players = self.manager.lookup_players_for_websocket(websocket)
        for player in players:
            self.handle_player_disconnected_event(player)
        self.manager.delete_websocket(websocket)

    def handle_websocket_idle_event(self, websocket: TrackedWebsocket) -> None:
        """Handle the Websocket Idle event."""
        log.info("Event - WEBSOCKET IDLE - %s", id(websocket.websocket))
        if websocket.activity_state != ActivityState.IDLE:
            message = Message(MessageType.WEBSOCKET_IDLE)
            self.queue.message(message, websockets=[websocket.websocket])
            websocket.mark_idle()

    def handle_websocket_inactive_event(self, websocket: TrackedWebsocket) -> None:
        """Handle the Websocket Inactive event."""
        log.info("Event - WEBSOCKET INACTIVE - %s", id(websocket.websocket))
        if websocket.activity_state != ActivityState.INACTIVE:
            websocket.mark_inactive()
            message = Message(MessageType.WEBSOCKET_INACTIVE)
            self.queue.message(message, websockets=[websocket.websocket])
            self.queue.disconnect(websocket.websocket)  # will eventually trigger handle_websocket_disconnected_event()

    def handle_registered_players_event(self, player: TrackedPlayer) -> None:
        """Handle the Registered Players event."""
        log.info("Event - REGISTERED PLAYERS - %s", player.handle)
        players = [player.to_registered_player() for player in self.manager.lookup_all_players()]
        context = RegisteredPlayersContext(players=players)
        message = Message(MessageType.REGISTERED_PLAYERS, context=context)
        self.queue.message(message, players=[player])

    def handle_available_games_event(self, player: TrackedPlayer) -> None:
        """Handle the Available Games event."""
        log.info("Event - AVAILABLE GAMES - %s", player.handle)
        games = [game.to_advertised_game() for game in self.manager.lookup_available_games(player)]
        context = AvailableGamesContext(games=games)
        message = Message(MessageType.AVAILABLE_GAMES, context=context)
        self.queue.message(message, players=[player])

    def handle_player_registered_event(self, websocket: WebSocketServerProtocol, handle: str) -> None:
        """Handle the Player Registered event."""
        log.info("Event - PLAYER REGISTERED - %s on %s", handle, id(websocket))
        player = self.manager.track_player(websocket, handle)
        context = PlayerRegisteredContext(handle=player.handle)
        message = Message(MessageType.PLAYER_REGISTERED, player_id=player.player_id, context=context)
        self.queue.message(message, websockets=[websocket])

    def handle_player_reregistered_event(self, player: TrackedPlayer, websocket: WebSocketServerProtocol) -> None:
        """Handle the Player Registered event."""
        log.info("Event - PLAYER REREGISTERED - %s on %s", player.handle, id(websocket))
        player.websocket = websocket
        context = PlayerRegisteredContext(handle=player.handle)
        message = Message(MessageType.PLAYER_REGISTERED, player_id=player.player_id, context=context)
        self.queue.message(message, players=[player])

    def handle_player_unregistered_event(self, player: TrackedPlayer, game: Optional[TrackedGame] = None) -> None:
        """Handle the Player Unregistered event."""
        log.info("Event - PLAYER UNREGISTERED - %s with game %s", player.handle, player.game_id)
        player.mark_quit()
        if game:
            comment = "Player %s unregistered" % player.handle
            self.handle_game_player_left_event(player, game, comment)
        self.manager.delete_player(player)

    def handle_player_disconnected_event(self, player: TrackedPlayer) -> None:
        """Handle the Player Disconnected event."""
        log.info("Event - PLAYER DISCONNECTED - %s", player.handle)
        game = self.manager.lookup_game(player=player)
        player.mark_disconnected()
        if game:
            comment = "Player %s disconnected" % player.handle
            self.handle_game_player_left_event(player, game, comment)

    def handle_player_idle_event(self, player: TrackedPlayer) -> None:
        """Handle the Player Idle event."""
        log.info("Event - PLAYER IDLE - %s", player.handle)
        if player.activity_state != ActivityState.IDLE:
            context = PlayerIdleContext(handle=player.handle)
            message = Message(MessageType.PLAYER_IDLE, context=context)
            self.queue.message(message, players=[player])
            player.mark_idle()

    def handle_player_inactive_event(self, player: TrackedPlayer) -> None:
        """Handle the Player Inactive event."""
        # Note that we do not disconnect the websocket, because it might still be in use by other players.
        # If this is the last registered player using the websocket, it will also eventually be marked inactive, and be closed then.
        log.info("Event - PLAYER INACTIVE - %s", player.handle)
        if player.activity_state != ActivityState.INACTIVE:
            player.mark_inactive()
            context = PlayerInactiveContext(handle=player.handle)
            message = Message(MessageType.PLAYER_INACTIVE, context=context)
            game = self.manager.lookup_game(player=player)
            self.queue.message(message, players=[player])
            self.handle_player_unregistered_event(player, game)

    def handle_player_message_received_event(self, sender_handle: str, recipient_handles: List[str], sender_message: str) -> None:
        """Handle the Player Message Received event."""
        log.info(
            "Event - PLAYER MESSAGE RECEIVED - %s sending %d bytes to %d recipients",
            sender_handle,
            len(sender_message),
            len(recipient_handles),
        )
        context = PlayerMessageReceivedContext(sender_handle, recipient_handles, sender_message)
        message = Message(MessageType.PLAYER_MESSAGE_RECEIVED, context=context)
        players = [self.manager.lookup_player(handle=handle) for handle in recipient_handles]
        self.queue.message(message, players=[player for player in players if player])

    def handle_game_advertised_event(self, player: TrackedPlayer, advertised: AdvertiseGameContext) -> None:
        """Handle the Game Advertised event."""
        log.info("Event - GAME ADVERTISED - %s", player.handle)
        game = self.manager.track_game(player, advertised)
        self.handle_game_joined_event(player, game=game)
        self.handle_game_invitation_event(game)
        context = GameAdvertisedContext(game=game.to_advertised_game())  # get result here so it shows advertiser as joined
        message = Message(MessageType.GAME_ADVERTISED, context=context)
        self.queue.message(message, players=[player])

    def handle_game_invitation_event(self, game: TrackedGame) -> None:
        """Handle the Game Invitation event."""
        log.info("Event - GAME INVITATION - %s", game.game_id)
        if game.invited_handles:
            context = GameInvitationContext(game=game.to_advertised_game())
            message = Message(MessageType.GAME_INVITATION, context=context)
            players = [self.manager.lookup_player(handle=handle) for handle in game.invited_handles]
            self.queue.message(message, players=[player for player in players if player])

    def handle_game_joined_event(
        self, player: TrackedPlayer, game_id: Optional[str] = None, game: Optional[TrackedGame] = None
    ) -> None:
        """Handle the Game Joined event."""
        log.info("Event - GAME JOINED - %s", game_id if game_id else game.game_id if game else None)
        if game_id:
            game = self.manager.lookup_game(game_id=game_id)
            if not game or not game.is_available(player.handle):
                raise ProcessingError(FailureReason.INVALID_GAME, handle=player.handle)
        if not game:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, comment="Invalid arguments", handle=player.handle)
        game.mark_active()
        player.mark_joined(game)
        game.mark_joined(player.handle)
        context = GameJoinedContext(game_id=game.game_id)
        message = Message(MessageType.GAME_JOINED, context=context)
        self.queue.message(message, players=[player])
        if game.is_fully_joined():
            if self.manager.get_in_progress_game_count() >= config().in_progress_game_limit:
                # Rather than giving the caller an error, we just ignore the game and force the
                # advertiser to manually start it sometime later.  At least then, if the limit
                # has still been reached, the player receiving the error will be able to make
                # sense of it.  It doesn't seem right to fail the join operation (which has
                # completed successfully by this point) because the game can't be started.
                log.warning("Game limit reached, so game %s will not be auto-started", game.game_id)
            else:
                self.handle_game_started_event(game)

    def handle_game_started_event(self, game: TrackedGame) -> None:
        """Handle the Game Started event."""
        log.info("Event - GAME STARTED - %s", game.game_id)
        context = GameStartedContext(game_id=game.game_id)
        message = Message(MessageType.GAME_STARTED, context=context)
        game.mark_active()
        game.mark_started()
        players = self.manager.lookup_game_players(game)
        for player in players:
            player.mark_playing()
        self.queue.message(message, players=players)
        self.handle_game_player_change_event(game, "Game started")
        self.handle_game_state_change_event(game)
        self.handle_game_next_turn_event(game)

    def handle_game_cancelled_event(
        self, game: TrackedGame, reason: CancelledReason, comment: Optional[str] = None, notify: bool = True
    ) -> None:
        """Handle the Game Cancelled event."""
        log.info("Event - GAME CANCELLED - %s for %s (%s)", game.game_id, reason, "'%s'" % comment if comment else None)
        context = GameCancelledContext(game_id=game.game_id, reason=reason, comment=comment)
        message = Message(MessageType.GAME_CANCELLED, context=context)
        players = self.manager.lookup_game_players(game)
        for player in players:
            player.mark_quit()
        game.mark_cancelled(reason, comment)
        if notify:
            self.queue.message(message, players=players)
            self.handle_game_state_change_event(game)

    def handle_game_completed_event(self, game: TrackedGame, comment: Optional[str] = None) -> None:
        """Handle the Game Completed event."""
        log.info("Event - GAME COMPLETED - %s (%s)", game.game_id, "'%s'" % comment if comment else None)
        context = GameCompletedContext(game_id=game.game_id, comment=comment)
        message = Message(MessageType.GAME_COMPLETED, context=context)
        players = self.manager.lookup_game_players(game)
        for player in players:
            player.mark_quit()
        game.mark_completed(comment)
        self.queue.message(message, players=players)
        self.handle_game_state_change_event(game)

    def handle_game_idle_event(self, game: TrackedGame) -> None:
        """Handle the Game Idle event."""
        log.info("Event - GAME IDLE - %s", game.game_id)
        if game.activity_state != ActivityState.IDLE:
            context = GameIdleContext(game_id=game.game_id)
            message = Message(MessageType.GAME_IDLE, context=context)
            players = self.manager.lookup_game_players(game)
            self.queue.message(message, players=players)

    def handle_game_inactive_event(self, game: TrackedGame) -> None:
        """Handle the Game Inactive event."""
        log.info("Event - GAME INACTIVE - %s", game.game_id)
        if game.activity_state != ActivityState.INACTIVE:
            game.mark_inactive()
            context = GameInactiveContext(game_id=game.game_id)
            message = Message(MessageType.GAME_INACTIVE, context=context)
            players = self.manager.lookup_game_players(game)
            self.queue.message(message, players=players)
            self.handle_game_cancelled_event(game, CancelledReason.INACTIVE)

    def handle_game_obsolete_event(self, game: TrackedGame) -> None:
        """Handle the Game Obsolete event."""
        log.info("Event - GAME OBSOLETE - %s", game.game_id)
        self.manager.delete_game(game)

    def handle_game_player_quit_event(self, player: TrackedPlayer, game: TrackedGame) -> None:
        """Handle the Game Player Quit event."""
        log.info("Event - GAME PLAYER QUIT - %s quit %s", player.handle, game.game_id)
        game.mark_active()
        player.mark_quit()
        context = GamePlayerQuitContext(handle=player.handle, game_id=game.game_id)
        message = Message(MessageType.GAME_PLAYER_QUIT, context=context)
        self.queue.message(message, players=[player])
        comment = "Player %s quit" % player.handle
        self.handle_game_player_left_event(player, game, comment)

    def handle_game_player_left_event(self, player: TrackedPlayer, game: TrackedGame, comment: str) -> None:
        """Handle the Game Player Left event."""
        log.info("Event - GAME PLAYER LEFT - %s left %s ('%s')", player.handle, game.game_id, comment)
        if player.handle == game.advertiser_handle:
            self.handle_game_cancelled_event(game, CancelledReason.CANCELLED, comment)
        else:
            game.mark_quit(player.handle)
            self.handle_game_player_change_event(game, comment)
            if not game.is_viable():
                self.handle_game_cancelled_event(game, CancelledReason.NOT_VIABLE, comment)
            elif game.is_playing() and game.is_move_pending(player.handle):
                # if the player is in the middle of their turn when they leave, we need to finish it for them
                self.handle_game_programmatic_move_event(player.handle, game)

    def handle_game_player_move_event(self, player: TrackedPlayer, game: TrackedGame, move_id: str) -> None:
        """Handle the Game Player Move event."""
        log.info("Event - GAME PLAYER MOVE - %s for %s, move %s", player.handle, game.game_id, move_id)
        self.handle_game_move_event(player.handle, game, move_id)

    def handle_game_programmatic_move_event(self, handle: str, game: TrackedGame) -> None:
        """Handle the Game Programmatic Move event."""
        log.info("Event - GAME PROGRAMMATIC MOVE - %s for %s", handle, game.game_id)
        view = game.get_player_view(handle)
        moves = game.get_legal_moves(handle)
        move = RewardV1InputSource().choose_move(game.mode, view, moves, Rules.evaluate_move)
        self.handle_game_move_event(handle, game, move.id)

    def handle_game_move_event(self, handle: str, game: TrackedGame, move_id: str) -> None:
        """Handle the Game Move event."""
        log.info("Event - GAME MOVE - %s for %s, move %s", handle, game.game_id, move_id)
        game.mark_active()
        (completed, comment) = game.execute_move(handle, move_id)
        if completed:
            self.handle_game_completed_event(game, comment)
        else:
            self.handle_game_state_change_event(game)
            self.handle_game_next_turn_event(game)

    def handle_game_next_turn_event(self, game: TrackedGame) -> None:
        """Handle the Game Next Turn event."""
        log.info("Event - GAME NEXT TURN - %s", game.game_id)
        handle, player_type = game.get_next_turn()
        if player_type == PlayerType.PROGRAMMATIC:
            self.handle_game_programmatic_move_event(handle, game)
        else:
            player = self.manager.lookup_player(handle=handle)
            if player and player.player_state == PlayerState.PLAYING:  # they might have quit or even have unregistered
                moves = game.get_legal_moves(handle)
                self.handle_game_player_turn_event(player, moves)
            else:
                # if the player is no longer available to play, then we execute a move programmatically
                self.handle_game_programmatic_move_event(handle, game)

    def handle_game_player_change_event(self, game: TrackedGame, comment: str) -> None:
        """Handle the Game Player Change event."""
        log.info("Event - GAME PLAYER CHANGE - %s (%s)", game.game_id, "'%s'" % comment if comment else None)
        players = self.manager.lookup_game_players(game)
        context = GamePlayerChangeContext(game_id=game.game_id, comment=comment, players=game.get_game_players())
        message = Message(MessageType.GAME_PLAYER_CHANGE, context=context)
        self.queue.message(message, players=players)

    # pylint: disable=redefined-argument-from-local
    def handle_game_state_change_event(self, game: TrackedGame, player: Optional[TrackedPlayer] = None) -> None:
        """Handle the Game State Change event."""
        log.info("Event - GAME STATE CHANGE - %s for %s", game.game_id, player.handle if player else None)
        game.mark_active()
        if game.is_playing():
            players = [player] if player else self.manager.lookup_game_players(game)
            for player in players:
                view = game.get_player_view(player.handle)
                history = game.get_recent_history(10)  # the last 10 entries in history
                context = GameStateChangeContext.for_context(game_id=game.game_id, view=view, history=history)
                message = Message(MessageType.GAME_STATE_CHANGE, context=context)
                self.queue.message(message, players=[player])

    def handle_game_player_turn_event(self, player: TrackedPlayer, moves: List[Move]) -> None:
        """Handle the Game Player Turn event."""
        log.info("Event - GAME PLAYER TURN - %s for %s (%d moves)", player.handle, player.game_id, len(moves))
        context = GamePlayerTurnContext.for_moves(handle=player.handle, game_id=player.game_id, moves=moves)  # type: ignore
        message = Message(MessageType.GAME_PLAYER_TURN, context=context)
        self.queue.message(message, players=[player])



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=wildcard-import

"""
Implements a quick'n'dirty game-playing client as a demo.

Because this is part of the apologiesserver codebase, it can use all
of the interface classes.  That greatly simplifies the implementation.
But, the entire demo takes place over a network connection, so it is
a real test of the websockets server.

See also: the TestGame class in test_manager.py, which implements a 
similar flow of control synchronously against the internal manager
interface.  If you need to debug something, that's probably easier
to work with.
"""

import asyncio
import logging
import random
import signal
from asyncio import AbstractEventLoop, CancelledError
from typing import List, Optional, Tuple, cast

import websockets
from apologies.game import GameMode
from websockets import WebSocketClientProtocol

from .interface import *
from .util import receive, send

SHUTDOWN_SIGNALS = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)

log = logging.getLogger("apologies.demo")


async def _register_player(websocket: WebSocketClientProtocol) -> str:
    """Register a player."""
    context = RegisterPlayerContext(handle="Demo")
    request = Message(MessageType.REGISTER_PLAYER, context=context)
    await send(websocket, request)
    response = await receive(websocket)
    log.info("Completed registering handle=leela, got player_id=%s", response.player_id)  # type: ignore
    return response.player_id  # type: ignore


async def _unregister_player(websocket: WebSocketClientProtocol, player_id: str) -> None:
    """Register a player."""
    request = Message(MessageType.UNREGISTER_PLAYER, player_id=player_id)
    await send(websocket, request)
    log.info("Completed unregistering handle=leela")


async def _advertise_game(websocket: WebSocketClientProtocol, player_id: str) -> None:
    """Advertise a game."""
    name = "Demo Game"
    mode = GameMode.STANDARD
    players = 4
    visibility = Visibility.PUBLIC
    invited_handles: List[str] = []
    context = AdvertiseGameContext(name, mode, players, visibility, invited_handles)
    request = Message(MessageType.ADVERTISE_GAME, player_id=player_id, context=context)
    await send(websocket, request)


async def _start_game(websocket: WebSocketClientProtocol, player_id: str) -> None:
    """Start the player's advertised game."""
    request = Message(MessageType.START_GAME, player_id=player_id)
    await send(websocket, request)


def _handle_game_joined(_player_id: str, message: Message) -> None:
    """Handle the game joined event."""
    context = cast(GameJoinedContext, message.context)
    log.info("Joined game %s", context.game_id)


def _handle_game_advertised(_player_id: str, message: Message) -> None:
    """Handle the game advertised event."""
    context = cast(GameAdvertisedContext, message.context)
    log.info("Advertised game '%s': %s", context.game.name, context.game.game_id)


def _handle_game_started(_player_id: str, message: Message) -> None:
    """Handle the game started event."""
    context = cast(GameStartedContext, message.context)
    log.info("Started game: %s", context.game_id)


def _handle_game_completed(_player_id: str, message: Message) -> None:
    """Handle the game completed event."""
    context = cast(GameCompletedContext, message.context)
    log.info("Game completed: %s", context.comment if context.comment else "")


def _handle_game_state_change(_player_id: str, message: Message) -> None:
    "Handle the game state change event"
    context = cast(GameStateChangeContext, message.context)
    if context.recent_history:
        history = context.recent_history[-1]
        color = "General" if not history.color else history.color.value
        action = history.action
        log.info("%s - %s", color, action)


def _handle_game_player_change(_player_id: str, message: Message) -> None:
    """Handle the game player change event."""
    context = cast(GamePlayerChangeContext, message.context)
    players = ["%s (%s)" % (player.handle, player.player_color.value) for player in context.players]  # type: ignore
    log.info("Game players are: %s", players)


def _handle_game_player_turn(player_id: str, message: Message) -> Optional[Message]:
    """Handle the game player turn event."""
    context = cast(GamePlayerTurnContext, message.context)
    move: GameMove = random.choice(list(context.moves.values()))
    log.info("Demo player turn, %d move(s), chose %s for card %s", len(context.moves), move.move_id, move.card.name)
    return Message(MessageType.EXECUTE_MOVE, player_id=player_id, context=ExecuteMoveContext(move_id=move.move_id))


def _handle_message(player_id: str, message: Message) -> Tuple[bool, Optional[Message]]:
    """Handle any message received from the connection."""
    completed = False
    response = None
    if message.message == MessageType.GAME_JOINED:
        _handle_game_joined(player_id, message)
    elif message.message == MessageType.GAME_ADVERTISED:
        _handle_game_advertised(player_id, message)
    elif message.message == MessageType.GAME_STARTED:
        _handle_game_started(player_id, message)
    elif message.message == MessageType.GAME_COMPLETED:
        _handle_game_completed(player_id, message)
        completed = True
    elif message.message == MessageType.GAME_STATE_CHANGE:
        _handle_game_state_change(player_id, message)
    elif message.message == MessageType.GAME_PLAYER_CHANGE:
        _handle_game_player_change(player_id, message)
    elif message.message == MessageType.GAME_PLAYER_TURN:
        response = _handle_game_player_turn(player_id, message)
    else:
        log.info("Ignored message with type %s", message.message.name)
    return completed, response


async def _handle_connection(websocket: WebSocketClientProtocol) -> None:
    """Handle a websocket connection, sending and receiving messages."""
    player_id = await _register_player(websocket)
    await _advertise_game(websocket, player_id)
    await _start_game(websocket, player_id)
    while True:
        message = await receive(websocket, timeout_sec=30)
        if message:
            completed, response = _handle_message(player_id, message)
            if response:
                await send(websocket, response)
            if completed:
                break
    await _unregister_player(websocket, player_id)


async def _websocket_client(uri: str) -> None:
    """The asynchronous websocket client."""
    log.info("Completed starting websocket client")  # ok, it's a bit of a lie
    try:
        async with websockets.connect(uri=uri) as websocket:
            await _handle_connection(websocket)
    except Exception as e:  # pylint: disable=broad-except
        log.error("Error with connection: %s", str(e), exc_info=True)


async def _terminate() -> None:
    """Terminate running tasks."""
    pending = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    for task in pending:
        task.cancel()
    await asyncio.gather(*pending)


def _add_signal_handlers(loop: AbstractEventLoop) -> None:
    """Add signal handlers so shutdown can be handled normally, returning the stop future."""
    log.info("Adding signal handlers...")
    for sig in SHUTDOWN_SIGNALS:
        loop.add_signal_handler(sig, lambda: asyncio.create_task(_terminate()))


def demo(host: str, port: int) -> None:
    """Run the demo."""
    uri = "ws://%s:%d" % (host, port)
    log.info("Demo client started against %s", uri)
    loop = asyncio.get_event_loop()
    _add_signal_handlers(loop)
    try:
        loop.run_until_complete(_websocket_client(uri))
    except CancelledError:
        pass
    finally:
        loop.stop()
        loop.close()
        log.info("Demo client finished")



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=wildcard-import,too-many-lines

"""
State manager.

Python's asyncio is primarily meant for use in single-threaded code, but there is still
concurrent execution happening any time we hit a yield from or await.

We want to minimize the risk of unexpected behavior when there are conflicting requests.
For instance, if we simultaneously get a request to start a game and to quit a game, we
want to make sure that one operation completes entirely before the next one starts.  This
means that we need thread synchronization whenever state is updated.

I've chosen to synchronize all state upate operations behind a single transaction boundary
(a single lock).  This is easier to follow and easier to write (correctly) than tracking
individual locks at a more granular level, like at the player or the game level.  The
state will never be locked for all that long, because state update operations are all done
in-memory and are quite fast.  The slow stuff like network requests all happen outside the
lock, whether we're processing a request or executing a scheduled task.

The design would be different if we were using a database to save state, but this seems
like the best compromise for the simple in-memory design that we're using now.

The transaction boundary is handled by a single lock on the StateManager object.  Callers
must ensure that they get that lock before reading or modifying state in any way.  Other 
than that, none of the objects defined in this module are thread-safe, or even thread-aware.  
There are no asynchronous methods or await calls.  This simplifies the implementation and 
avoids confusion.
"""

from __future__ import annotations  # see: https://stackoverflow.com/a/33533514/2907667

import asyncio
import logging
import random
from typing import Dict, List, Optional, Tuple
from uuid import uuid4

import attr
import pendulum
from apologies.engine import Character, Engine
from apologies.game import GameMode, History, PlayerColor, PlayerView
from apologies.rules import Move
from apologies.source import NoOpInputSource
from ordered_set import OrderedSet  # this makes expected results easier to articulate in test code
from pendulum.datetime import DateTime
from websockets import WebSocketServerProtocol

from .interface import *

log = logging.getLogger("apologies.manager")


# These are names we can assign to programmatic players.
_NAMES = [
    "Aragorn",
    "Arwen",
    "Bilbo",
    "Boromir",
    "Elrond",
    "Éomer",
    "Éowyn",
    "Faramir",
    "Frodo",
    "Galadriel",
    "Gandalf",
    "Gimli",
    "Gollum",
    "Isildur",
    "Legolas",
    "Merry",
    "Pippen",
    "Radagast",
    "Samwise",
    "Saruman",
    "Sauron",
    "Shelob",
    "Théoden",
    "Treebeard",
]


@attr.s
class TrackedWebsocket:
    """The state that is tracked for a websocket within the state manager."""

    websocket = attr.ib(type=WebSocketServerProtocol)
    registration_date = attr.ib(type=DateTime)
    last_active_date = attr.ib(type=DateTime)
    activity_state = attr.ib(type=ActivityState, default=ActivityState.ACTIVE)
    player_ids = attr.ib(type=OrderedSet[str])

    @registration_date.default
    def _default_registration_date(self) -> DateTime:
        return pendulum.now()

    @last_active_date.default
    def _default_last_active_date(self) -> DateTime:
        return pendulum.now()

    @player_ids.default
    def _default_player_ids(self) -> OrderedSet[str]:
        return OrderedSet()

    def mark_active(self) -> None:
        """Mark the websocket as active."""
        self.last_active_date = pendulum.now()
        self.activity_state = ActivityState.ACTIVE

    def mark_idle(self) -> None:
        """Mark the websocket as idle."""
        self.activity_state = ActivityState.IDLE

    def mark_inactive(self) -> None:
        """Mark the websocket as inactive."""
        self.activity_state = ActivityState.INACTIVE


# pylint: disable=too-many-instance-attributes
@attr.s
class TrackedPlayer:
    """The state that is tracked for a player within the state manager."""

    player_id = attr.ib(type=str, repr=False)  # treat as read-only; this is a secret, so we don't want it printed or logged
    handle = attr.ib(type=str)  # treat as read-only
    websocket = attr.ib(type=Optional[WebSocketServerProtocol])
    registration_date = attr.ib(type=DateTime)
    last_active_date = attr.ib(type=DateTime)
    activity_state = attr.ib(type=ActivityState, default=ActivityState.ACTIVE)
    connection_state = attr.ib(type=ConnectionState, default=ConnectionState.CONNECTED)
    player_state = attr.ib(type=PlayerState, default=PlayerState.WAITING)
    game_id = attr.ib(type=Optional[str], default=None)

    @registration_date.default
    def _default_registration_date(self) -> DateTime:
        return pendulum.now()

    @last_active_date.default
    def _default_last_active_date(self) -> DateTime:
        return pendulum.now()

    @staticmethod
    def for_context(player_id: str, websocket: WebSocketServerProtocol, handle: str) -> TrackedPlayer:
        """Create a tracked player based on provided context."""
        return TrackedPlayer(player_id=player_id, websocket=websocket, handle=handle)

    def to_registered_player(self) -> RegisteredPlayer:
        """Convert this TrackedPlayer to a RegisteredPlayer."""
        return RegisteredPlayer(
            handle=self.handle,
            registration_date=self.registration_date,
            last_active_date=self.last_active_date,
            connection_state=self.connection_state,
            activity_state=self.activity_state,
            player_state=self.player_state,
            game_id=self.game_id,
        )

    def mark_active(self) -> None:
        """Mark the player as active."""
        self.last_active_date = pendulum.now()
        self.activity_state = ActivityState.ACTIVE
        self.connection_state = ConnectionState.CONNECTED

    def mark_idle(self) -> None:
        """Mark the player as idle."""
        self.activity_state = ActivityState.IDLE

    def mark_inactive(self) -> None:
        """Mark the player as inactive."""
        self.activity_state = ActivityState.INACTIVE

    def mark_joined(self, game: TrackedGame) -> None:
        """Mark that the player has joined a game."""
        self.game_id = game.game_id
        self.player_state = PlayerState.JOINED

    def mark_playing(self) -> None:
        """Mark that the player is playing a game."""
        self.player_state = PlayerState.PLAYING

    def mark_quit(self) -> None:
        """Mark that the player has quit a game."""
        self.game_id = None
        self.player_state = PlayerState.WAITING  # they go right through QUIT and back to WAITING

    def mark_disconnected(self) -> None:
        """Mark the player as disconnected."""
        self.mark_quit()
        self.websocket = None
        self.activity_state = ActivityState.IDLE
        self.connection_state = ConnectionState.DISCONNECTED


@attr.s(frozen=True)
class CurrentTurn:

    handle = attr.ib(type=str)
    color = attr.ib(type=PlayerColor)
    view = attr.ib(type=PlayerView)
    movelist = attr.ib(type=List[Move])
    movedict = attr.ib(type=Dict[str, Move])

    def draw_again(self, engine: Engine) -> CurrentTurn:
        return CurrentTurn.for_handle(engine, self.handle, self.color)

    @staticmethod
    def next_player(engine: Engine) -> CurrentTurn:
        color, character = engine.next_turn()
        return CurrentTurn.for_handle(engine, character.name, color)

    @staticmethod
    def for_handle(engine: Engine, handle: str, color: PlayerColor) -> CurrentTurn:
        view = engine.game.create_player_view(color)
        _, movelist = engine.construct_legal_moves(view)
        movedict = {move.id: move for move in movelist}
        return CurrentTurn(handle=handle, color=color, view=view, movelist=movelist, movedict=movedict)


@attr.s
class TrackedEngine:
    """Wrapper over an Apologies game engine, to manage game play state for TrackedGame."""

    _engine = attr.ib(type=Optional[Engine], default=None)
    _colors = attr.ib(type=Dict[str, PlayerColor])
    _current = attr.ib(type=Optional[CurrentTurn], default=None)

    @_colors.default
    def _default_colors(self) -> Dict[str, PlayerColor]:
        return {}

    def start_game(self, mode: GameMode, handles: List[str]) -> Dict[str, PlayerColor]:
        """Start the game, returning a map from handle to assigned color."""
        if self._engine:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        characters = [Character(name=handle, source=NoOpInputSource()) for handle in handles]
        self._engine = Engine(mode, characters=characters)
        self._engine.start_game()
        self._colors = {character.name: color for color, character in self._engine.colors.items()}
        self._current = CurrentTurn.next_player(self._engine)
        return self._colors.copy()  # copy so caller can't mess with internal state

    def stop_game(self) -> None:
        """Stop the game after the game is completed or has been cancelled."""
        self._engine = None
        self._colors = {}
        self._current = None

    def get_next_turn(self) -> str:
        """Get the handle of the player that should play the next turn."""
        if not self._current:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        return self._current.handle

    def get_legal_moves(self, handle: str) -> List[Move]:
        """Get the legal moves for the player at this stage in the game."""
        if not self._current or handle != self._current.handle:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        return self._current.movelist[:]

    def get_player_view(self, handle: str) -> PlayerView:
        """Get the player's view of the game state."""
        if not self._engine or handle not in self._colors:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        color = self._colors[handle]
        return self._engine.game.create_player_view(color)

    def get_recent_history(self, max_entries: int) -> List[History]:
        """Return up to a certain number of game history entries."""
        if not self._engine or not self._engine.game:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        return self._engine.game.history[-max_entries:]

    def is_move_pending(self, handle: str) -> bool:
        """Whether a move is pending for the player with the passed-in handle."""
        if not self._current:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        return handle == self._current.handle

    def is_legal_move(self, handle: str, move_id: str) -> bool:
        """Whether the passed-in move id is a legal move for the player."""
        if not self._current:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        return handle == self._current.handle and move_id in self._current.movedict

    def execute_move(self, handle: str, move_id: str) -> Tuple[bool, Optional[str]]:
        """Execute a player's move, returning whether the game was completed (and a comment if so)."""
        if (
            not self._engine
            or not self._current
            or not handle == self._current.handle
            or handle not in self._colors
            or move_id not in self._current.movedict
        ):
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        color = self._colors[handle]
        move = self._current.movedict[move_id]
        done = self._engine.execute_move(color, move)
        if self._engine.completed:
            self._current = None
            character, player = self._engine.winner()  # type: ignore
            comment = "Player %s (%s) won after %d turns" % (character.name, player.color.name, player.turns)
            return True, comment
        self._current = CurrentTurn.next_player(self._engine) if done else self._current.draw_again(self._engine)
        return False, None


# pylint: disable=too-many-instance-attributes,too-many-public-methods
@attr.s
class TrackedGame:
    """The state that is tracked for a game within the state manager."""

    game_id = attr.ib(type=str)  # treat as read-only
    advertiser_handle = attr.ib(type=str)  # treat as read-only
    name = attr.ib(type=str)  # treat as read-only
    mode = attr.ib(type=GameMode)  # treat as read-only
    players = attr.ib(type=int)  # treat as read-only
    visibility = attr.ib(type=Visibility)  # treat as read-only
    invited_handles = attr.ib(type=List[str])  # treat as read-only
    advertised_date = attr.ib(type=DateTime)
    last_active_date = attr.ib(type=DateTime)
    started_date = attr.ib(type=Optional[DateTime], default=None)
    completed_date = attr.ib(type=Optional[DateTime], default=None)
    game_state = attr.ib(type=GameState, default=GameState.ADVERTISED)
    activity_state = attr.ib(type=ActivityState, default=ActivityState.ACTIVE)
    cancelled_reason = attr.ib(type=Optional[CancelledReason], default=None)
    completed_comment = attr.ib(type=Optional[str], default=None)
    game_players = attr.ib(type=Dict[str, GamePlayer])
    _engine = attr.ib(type=TrackedEngine)

    @advertised_date.default
    def _default_advertised_date(self) -> DateTime:
        return pendulum.now()

    @last_active_date.default
    def _default_last_active_date(self) -> DateTime:
        return pendulum.now()

    @game_players.default
    def _default_game_players(self) -> Dict[str, GamePlayer]:
        return {}

    @_engine.default
    def _default_engine(self) -> TrackedEngine:
        return TrackedEngine()

    @staticmethod
    def for_context(advertiser_handle: str, game_id: str, context: AdvertiseGameContext) -> TrackedGame:
        """Create a tracked game based on provided context."""
        return TrackedGame(
            game_id=game_id,
            advertiser_handle=advertiser_handle,
            name=context.name,
            mode=context.mode,
            players=context.players,
            visibility=context.visibility,
            invited_handles=context.invited_handles[:],
        )

    def to_advertised_game(self) -> AdvertisedGame:
        """Convert this tracked game to an AdvertisedGame."""
        return AdvertisedGame(
            game_id=self.game_id,
            name=self.name,
            mode=self.mode,
            advertiser_handle=self.advertiser_handle,
            players=self.players,
            available=self.players - len(self.game_players),
            visibility=self.visibility,
            invited_handles=self.invited_handles[:],
        )

    def get_game_players(self) -> List[GamePlayer]:
        """Get a list of game players."""
        return list(self.game_players.values())

    def get_available_players(self) -> List[GamePlayer]:
        """Get the players that are still available to play the game."""
        return [
            player for player in self.get_game_players() if player.player_state not in (PlayerState.QUIT, PlayerState.DISCONNECTED)
        ]

    def get_available_player_count(self) -> int:
        """Get the number of players that are still available to play the game."""
        return len(self.get_available_players())

    def get_next_turn(self) -> Tuple[str, PlayerType]:
        """Get the next turn to be played."""
        handle = self._engine.get_next_turn()
        player = self.game_players[handle]
        return handle, player.player_type

    def get_legal_moves(self, handle: str) -> List[Move]:
        """Get the legal moves for the player at this stage in the game."""
        return self._engine.get_legal_moves(handle)

    def get_player_view(self, handle: str) -> PlayerView:
        """Get the player's view of the game state."""
        return self._engine.get_player_view(handle)

    def get_recent_history(self, max_entries: int) -> List[History]:
        """Return up to a certain number of game history entries."""
        return self._engine.get_recent_history(max_entries)

    def is_available(self, handle: str) -> bool:
        """Whether the game is available to be joined by the passed-in player."""
        return self.game_state == GameState.ADVERTISED and (self.visibility == Visibility.PUBLIC or handle in self.invited_handles)

    def is_advertised(self) -> bool:
        """Whether a game is currently being advertised."""
        return self.game_state == GameState.ADVERTISED

    def is_playing(self) -> bool:
        """Whether a game is being played."""
        return self.game_state == GameState.PLAYING

    def is_in_progress(self) -> bool:
        """Whether a game is in-progress, meaning it is advertised or being played."""
        return self.is_advertised() or self.is_playing()

    def is_fully_joined(self) -> bool:
        """Whether the number of requested players have joined the game."""
        return self.players == len(self.game_players)

    def is_viable(self) -> bool:
        """Whether the game is viable."""
        # Advertised games are always viable. Otherwise, the game is only viable if at least two players remain to play turns.
        if self.is_advertised():
            return True
        else:
            return self.get_available_player_count() >= 2

    def is_move_pending(self, handle: str) -> bool:
        """Whether a move is pending for the player with the passed-in handle."""
        return self._engine.is_move_pending(handle)

    def is_legal_move(self, handle: str, move_id: str) -> bool:
        """Whether the passed-in move id is a legal move for the player."""
        return self._engine.is_legal_move(handle, move_id)

    def mark_active(self) -> None:
        """Mark the game as active."""
        self.last_active_date = pendulum.now()
        self.activity_state = ActivityState.ACTIVE

    def mark_idle(self) -> None:
        """Mark the game as idle."""
        self.activity_state = ActivityState.IDLE

    def mark_inactive(self) -> None:
        """Mark the game as inactive."""
        self.activity_state = ActivityState.INACTIVE

    def mark_joined(self, handle: str) -> None:
        """Mark that a player has joined a game."""
        if self.game_state != GameState.ADVERTISED:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        player_color = None  # will be assigned after game starts
        player_type = PlayerType.HUMAN
        player_state = PlayerState.JOINED
        self.game_players[handle] = GamePlayer(handle, player_color, player_type, player_state)

    def mark_started(self) -> None:
        """Mark the game as started."""
        if self.game_state != GameState.ADVERTISED:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        self.game_state = GameState.PLAYING
        self.last_active_date = pendulum.now()
        self.started_date = pendulum.now()
        for _ in range(0, self.players - len(self.game_players)):
            self._mark_joined_programmatic()  # fill in remaining players as necessary
        colors = self._engine.start_game(self.mode, list(self.game_players.keys()))
        for handle in self.game_players:
            self.game_players[handle] = attr.evolve(
                self.game_players[handle], player_color=colors[handle], player_state=PlayerState.PLAYING
            )

    def mark_completed(self, comment: Optional[str]) -> None:
        """Mark the game as completed."""
        if self.game_state != GameState.PLAYING:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        self.completed_date = pendulum.now()
        self.game_state = GameState.COMPLETED
        self.completed_comment = comment
        for handle in self.game_players:
            self.game_players[handle] = attr.evolve(self.game_players[handle], player_state=PlayerState.FINISHED)
        self._engine.stop_game()

    def mark_cancelled(self, reason: CancelledReason, comment: Optional[str] = None) -> None:
        """Mark the game as cancelled."""
        if self.game_state not in [GameState.ADVERTISED, GameState.PLAYING]:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
        self.completed_date = pendulum.now()
        self.game_state = GameState.CANCELLED
        self.cancelled_reason = reason
        self.completed_comment = comment
        for handle in self.game_players:
            self.game_players[handle] = attr.evolve(self.game_players[handle], player_state=PlayerState.FINISHED)
        self._engine.stop_game()

    def mark_quit(self, handle: str) -> None:
        """Mark that the player has quit a game."""
        # We assume that if the player is in the middle of their turn, that the caller handles that cleanup
        if self.game_state == GameState.ADVERTISED:
            del self.game_players[handle]  # if the game hasn't started, just remove them
        elif self.game_state == GameState.PLAYING:
            self.game_players[handle] = attr.evolve(self.game_players[handle], player_state=PlayerState.QUIT)
        else:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")

    def execute_move(self, handle: str, move_id: str) -> Tuple[bool, Optional[str]]:
        """Execute a player's move, returning an indication of whether the game was completed (and a comment if so)."""
        return self._engine.execute_move(handle, move_id)

    def _mark_joined_programmatic(self) -> None:
        """Create an join a programmatic player, assumed to be called from within a lock."""
        handle = self._assign_handle()
        player_color = None  # will be assigned after game starts
        player_type = PlayerType.PROGRAMMATIC
        player_state = PlayerState.JOINED
        self.game_players[handle] = GamePlayer(handle, player_color, player_type, player_state)

    def _assign_handle(self) -> str:
        """Assign a handle to the newly-joined programmatic player, assumed to be called from within a lock."""
        all_handles = set(_NAMES)
        used_handles = {player.handle for player in self.game_players.values()}
        available_handles = all_handles - used_handles
        return random.choice(list(available_handles))


# noinspection PyMethodMayBeStatic
@attr.s
class StateManager:

    """Manages system state."""

    lock = attr.ib(type=asyncio.Lock)
    _websocket_map = attr.ib(type=Dict[WebSocketServerProtocol, TrackedWebsocket])
    _game_map = attr.ib(type=Dict[str, TrackedGame])
    _player_map = attr.ib(type=Dict[str, TrackedPlayer])
    _handle_map = attr.ib(type=Dict[str, str])

    @lock.default
    def _default_lock(self) -> asyncio.Lock:
        return asyncio.Lock()

    @_websocket_map.default
    def _default_websocket_map(self) -> Dict[WebSocketServerProtocol, TrackedWebsocket]:
        return {}

    @_game_map.default
    def _default_game_map(self) -> Dict[str, TrackedGame]:
        return {}

    @_player_map.default
    def _default_player_map(self) -> Dict[str, TrackedPlayer]:
        return {}

    @_handle_map.default
    def _default_handle_map(self) -> Dict[str, str]:
        return {}

    def mark_active(self, player: TrackedPlayer) -> None:
        """Mark a player and its associated websocket as active."""
        if not player.websocket or player.websocket not in self._websocket_map:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, comment="Did not find player websocket", handle=player.handle)
        websocket = self._websocket_map[player.websocket]
        websocket.mark_active()
        player.mark_active()

    def track_websocket(self, websocket: WebSocketServerProtocol) -> None:
        """Track a connected websocket."""
        if websocket in self._websocket_map:
            raise ProcessingError(FailureReason.INTERNAL_ERROR, comment="Duplicate websocket encountered")
        self._websocket_map[websocket] = TrackedWebsocket(websocket=websocket)

    def delete_websocket(self, websocket: WebSocketServerProtocol) -> None:
        """Delete a websocket, so it is no longer tracked."""
        if websocket in self._websocket_map:
            del self._websocket_map[websocket]

    def get_websocket_count(self) -> int:
        """Return the number of connected websockets in the system."""
        return len(self._websocket_map)

    def lookup_all_websockets(self) -> List[WebSocketServerProtocol]:
        """Return a list of websockets for all tracked players."""
        return list(self._websocket_map.keys())

    def lookup_players_for_websocket(self, websocket: WebSocketServerProtocol) -> List[TrackedPlayer]:
        """Look up the players associated with a websocket, if any."""
        players = []
        if websocket in self._websocket_map:
            players = [self.lookup_player(player_id=player_id) for player_id in self._websocket_map[websocket].player_ids]
        return [player for player in players if player is not None]

    def track_game(self, player: TrackedPlayer, advertised: AdvertiseGameContext) -> TrackedGame:
        """Track a newly-advertised game."""
        game_id = "%s" % uuid4()
        self._game_map[game_id] = TrackedGame.for_context(player.handle, game_id, advertised)
        return self._game_map[game_id]

    def delete_game(self, game: TrackedGame) -> None:
        """Delete a tracked game, so it is no longer tracked."""
        if game.game_id in self._game_map:
            del self._game_map[game.game_id]

    def get_total_game_count(self) -> int:
        """Return the total number of games in the system."""
        return len(self._game_map)

    def get_in_progress_game_count(self) -> int:
        """Return the number of in-progress games in the system."""
        return len([game for game in self._game_map.values() if game.is_in_progress()])

    def lookup_game(self, game_id: Optional[str] = None, player: Optional[TrackedPlayer] = None) -> Optional[TrackedGame]:
        """Look up a game by id, returning None if the game is not found."""
        if game_id:
            return self._game_map[game_id] if game_id in self._game_map else None
        elif player:
            return self.lookup_game(game_id=player.game_id)
        else:
            return None

    def lookup_all_games(self) -> List[TrackedGame]:
        """Return a list of all tracked games."""
        return list(self._game_map.values())

    def lookup_in_progress_games(self) -> List[TrackedGame]:
        """Return a list of all in-progress games."""
        return [game for game in self.lookup_all_games() if game.is_in_progress()]

    def lookup_game_players(self, game: TrackedGame) -> List[TrackedPlayer]:
        """Lookup the players that are currently playing a game."""
        handles = [player.handle for player in game.game_players.values() if player.player_type == PlayerType.HUMAN]
        players = [self.lookup_player(handle=handle) for handle in handles]
        return [player for player in players if player is not None]

    def lookup_available_games(self, player: TrackedPlayer) -> List[TrackedGame]:
        """Return a list of games the passed-in player may join."""
        games = self.lookup_all_games()
        return [game for game in games if game.is_available(player.handle)]

    def track_player(self, websocket: WebSocketServerProtocol, handle: str) -> TrackedPlayer:
        """Track a newly-registered player."""
        if handle in self._handle_map:
            raise ProcessingError(FailureReason.DUPLICATE_USER, handle=handle)
        player_id = "%s" % uuid4()
        self._websocket_map[websocket].player_ids.add(player_id)
        self._player_map[player_id] = TrackedPlayer.for_context(player_id, websocket, handle)
        self._handle_map[handle] = player_id
        return self._player_map[player_id]

    def delete_player(self, player: TrackedPlayer) -> None:
        """Delete a tracked player, so it is no longer tracked."""
        if player.websocket and player.websocket in self._websocket_map:
            self._websocket_map[player.websocket].player_ids.discard(player.player_id)
        if player.handle in self._handle_map:
            del self._handle_map[player.handle]
        if player.player_id in self._player_map:
            del self._player_map[player.player_id]

    def get_registered_player_count(self) -> int:
        """Return the number of registered players in the system."""
        return len(self._player_map)

    def lookup_player(self, player_id: Optional[str] = None, handle: Optional[str] = None) -> Optional[TrackedPlayer]:
        """Look up a player by either player id or handle."""
        if player_id:
            return self._player_map[player_id] if player_id in self._player_map else None
        elif handle:
            player_id = self._handle_map[handle] if handle in self._handle_map else None
            return self.lookup_player(player_id=player_id)
        else:
            return None

    def lookup_all_players(self) -> List[TrackedPlayer]:
        """Return a list of all tracked players."""
        return list(self._player_map.values())

    def lookup_websocket_activity(self) -> List[Tuple[TrackedWebsocket, DateTime, int]]:
        """Look up the last active date and number of registered players for all websockets."""
        result: List[Tuple[TrackedWebsocket, DateTime, int]] = []
        for websocket in self._websocket_map.values():
            result.append((websocket, websocket.last_active_date, len(websocket.player_ids)))
        return result

    def lookup_player_activity(self) -> List[Tuple[TrackedPlayer, DateTime, ConnectionState]]:
        """Look up the last active date and connection state for all players."""
        result: List[Tuple[TrackedPlayer, DateTime, ConnectionState]] = []
        for player in self._player_map.values():
            result.append((player, player.last_active_date, player.connection_state))
        return result

    def lookup_game_activity(self) -> List[Tuple[TrackedGame, DateTime]]:
        """Look up the last active date for all games."""
        result: List[Tuple[TrackedGame, DateTime]] = []
        for game in self._game_map.values():
            result.append((game, game.last_active_date))
        return result

    def lookup_game_completion(self) -> List[Tuple[TrackedGame, Optional[DateTime]]]:
        """Look up the completed date for all completed games."""
        result: List[Tuple[TrackedGame, Optional[DateTime]]] = []
        for game in self._game_map.values():
            if game.game_state == GameState.COMPLETED:
                result.append((game, game.completed_date))
        return result


_MANAGER = StateManager()


def manager() -> StateManager:
    """Return the state manager."""
    return _MANAGER



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

import asyncio
import logging
import signal
from asyncio import AbstractEventLoop, Future  # pylint: disable=unused-import
from typing import Any, Callable, Coroutine, Union  # pylint: disable=unused-import

import websockets
from websockets import WebSocketServerProtocol

from .config import config
from .event import EventHandler, RequestContext
from .interface import FailureReason, Message, MessageType, ProcessingError, RequestFailedContext
from .manager import manager
from .scheduled import scheduled_tasks
from .util import close, mask, send

log = logging.getLogger("apologies.server")

SHUTDOWN_SIGNALS = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)

# pylint: disable=too-many-return-statements,too-many-branches
def _lookup_method(handler: EventHandler, message: MessageType) -> Callable[[RequestContext], None]:
    """Lookup the handler method to invoke for a message type."""
    if message == MessageType.REREGISTER_PLAYER:
        return handler.handle_reregister_player_request
    elif message == MessageType.UNREGISTER_PLAYER:
        return handler.handle_unregister_player_request
    elif message == MessageType.LIST_PLAYERS:
        return handler.handle_list_players_request
    elif message == MessageType.ADVERTISE_GAME:
        return handler.handle_advertise_game_request
    elif message == MessageType.LIST_AVAILABLE_GAMES:
        return handler.handle_list_available_games_request
    elif message == MessageType.JOIN_GAME:
        return handler.handle_join_game_request
    elif message == MessageType.QUIT_GAME:
        return handler.handle_quit_game_request
    elif message == MessageType.START_GAME:
        return handler.handle_start_game_request
    elif message == MessageType.CANCEL_GAME:
        return handler.handle_cancel_game_request
    elif message == MessageType.EXECUTE_MOVE:
        return handler.handle_execute_move_request
    elif message == MessageType.RETRIEVE_GAME_STATE:
        return handler.handle_retrieve_game_state_request
    elif message == MessageType.SEND_MESSAGE:
        return handler.handle_send_message_request
    else:
        raise ProcessingError(FailureReason.INTERNAL_ERROR, comment="Invalid request %s" % message.name)


def _handle_message(handler: EventHandler, message: Message, websocket: WebSocketServerProtocol) -> None:
    """Handle a valid message received from a websocket client, assumed to be within the handler's lock."""
    if message.message == MessageType.REGISTER_PLAYER:
        handler.handle_register_player_request(message, websocket)
    else:
        player = handler.manager.lookup_player(player_id=message.player_id)
        if not player:
            raise ProcessingError(FailureReason.INVALID_PLAYER)
        handler.manager.mark_active(player)  # marks both the player and its websocket as active
        log.debug("Request is for player: %s", player)
        game = handler.manager.lookup_game(game_id=player.game_id)
        request = RequestContext(message, websocket, player, game)
        method = _lookup_method(handler, request.message.message)
        method(request)


async def _handle_data(data: Union[str, bytes], websocket: WebSocketServerProtocol) -> None:
    """Handle data received from a websocket client."""
    log.debug("Received raw data for websocket %s:\n%s", id(websocket), mask(data))
    message = Message.for_json(str(data))
    with EventHandler(manager()) as handler:
        async with handler.manager.lock:
            _handle_message(handler, message, websocket)
        await handler.execute_tasks()


async def _handle_connect(websocket: WebSocketServerProtocol) -> None:
    """Handle a newly-connected client."""
    with EventHandler(manager()) as handler:
        async with handler.manager.lock:
            handler.handle_websocket_connected_event(websocket)
        await handler.execute_tasks()


async def _handle_disconnect(websocket: WebSocketServerProtocol) -> None:
    """Handle a disconnected client."""
    with EventHandler(manager()) as handler:
        async with handler.manager.lock:
            handler.handle_websocket_disconnected_event(websocket)
        await handler.execute_tasks()


# pylint: disable=broad-except
# noinspection PyBroadException
async def _handle_exception(exception: Exception, websocket: WebSocketServerProtocol) -> None:
    """Handle an exception by sending a request failed event."""
    try:
        disconnect = False
        try:
            log.error("Handling exception: %s", str(exception), exc_info=True)
            raise exception
        except ProcessingError as e:
            disconnect = e.reason == FailureReason.WEBSOCKET_LIMIT  # this is a special case that can't easily be handled elsewhere
            reason = e.reason
            comment = e.comment if e.comment else e.reason.value
            handle = e.handle
            context = RequestFailedContext(reason=reason, comment=comment, handle=handle)
        except ValueError as e:
            context = RequestFailedContext(FailureReason.INVALID_REQUEST, str(e))
        except Exception as e:
            # Note: we don't want to expose internal implementation details in the case of an internal error
            context = RequestFailedContext(FailureReason.INTERNAL_ERROR, FailureReason.INTERNAL_ERROR.value)
        message = Message(MessageType.REQUEST_FAILED, context=context)
        await send(websocket, message.to_json())
        if disconnect:
            await close(websocket)
    except Exception as e:
        # We don't propogate errors like this to the caller.  We just ignore them and
        # hope that we can recover for future requests.  If the websocket is dead,
        # we presume (hope?) that it will eventually get disconnected by the library.
        # It's not entirely clear what would be a better approach.  The only other
        # obvious option is to drop the client because we failed to send them an error,
        # which doesn't seem quite right, either.
        log.error("Failed to handle exception: %s", str(e))


# pylint: disable=broad-except
# noinspection PyBroadException
async def _handle_connection(websocket: WebSocketServerProtocol, _path: str) -> None:
    """Handle a client connection, invoked once for each client that connects to the server."""
    await _handle_connect(websocket)
    try:
        async for data in websocket:
            try:
                await _handle_data(data, websocket)
            except Exception as e:
                await _handle_exception(e, websocket)
    except websockets.exceptions.ConnectionClosed:  # we get this if the connection closes for any reason
        pass
    await _handle_disconnect(websocket)


async def _handle_shutdown() -> None:
    """Handle server shutdown."""
    with EventHandler(manager()) as handler:
        async with handler.manager.lock:
            handler.handle_server_shutdown_event()
        await handler.execute_tasks()


async def _websocket_server(stop: "Future[Any]", host: str, port: int, close_timeout_sec: float) -> None:
    """Websocket server."""
    log.info("Completed starting websocket server")  # ok, it's a bit of a lie
    async with websockets.serve(_handle_connection, host=host, port=port, close_timeout=close_timeout_sec):
        await stop
        await _handle_shutdown()


def _add_signal_handlers(loop: AbstractEventLoop) -> "Future[Any]":
    """Add signal handlers so shutdown can be handled normally, returning the stop future."""
    log.info("Adding signal handlers...")
    stop = loop.create_future()
    for sig in SHUTDOWN_SIGNALS:
        loop.add_signal_handler(sig, stop.set_result, None)
    return stop


def _schedule_tasks(loop: AbstractEventLoop) -> None:
    """Schedule all of the scheduled tasks."""
    log.info("Scheduling tasks...")
    for task in scheduled_tasks():
        loop.create_task(task())


def _run_server(loop: AbstractEventLoop, stop: "Future[Any]") -> None:

    """Run the websocket server, stopping and closing the event loop when the server completes."""

    # Websockets has this to say about closing connections:
    #
    #    The close_timeout parameter defines a maximum wait time in seconds for
    #    completing the closing handshake and terminating the TCP connection.
    #    close() completes in at most 4 * close_timeout on the server side and
    #    5 * close_timeout on the client side.
    #
    # Since our configuration specifies a maximum time to wait (in total), we
    # need to divide by 4 to make sure that time isn't exceeded.
    #
    # See also: https://websockets.readthedocs.io/en/stable/api.html#module-websockets.protocol
    #           https://websockets.readthedocs.io/en/stable/api.html#module-websockets.server

    host = config().server_host
    port = config().server_port
    close_timeout_sec = config().close_timeout_sec / 4
    loop.run_until_complete(_websocket_server(stop=stop, host=host, port=port, close_timeout_sec=close_timeout_sec))
    loop.stop()
    loop.close()


def server() -> None:
    """The main processing loop for the websockets server."""
    log.info("Apologies server started")
    log.info("Configuration: %s", config().to_json())
    loop = asyncio.get_event_loop()
    stop = _add_signal_handlers(loop)
    _schedule_tasks(loop)
    _run_server(loop, stop)
    log.info("Apologies server finished")



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

import argparse
import sys
from typing import Any, List

from .config import DEFAULT_CONFIG_PATH, DEFAULT_SERVER_HOST, DEFAULT_SERVER_PORT, config, load_config
from .demo import demo
from .server import server
from .util import setup_logging


def run_server(argv: List[str]) -> None:
    """Start the Apologies server."""

    parser = argparse.ArgumentParser(
        description="Start the apologies server and let it run forever.",
        epilog="By default, the server writes logs to stdout. If you prefer, you can "
        "specify the path to a logfile, and logs will be written there instead.  "
        'The default configuration file is "%s".  '
        "If the default configuration file is not found, default values will be set.  "
        "If you override the default config file, it must exist.  "
        'You may override any individual config parameter with "--override param:value".' % DEFAULT_CONFIG_PATH,
    )

    parser.add_argument("--quiet", action="store_true", help="decrease log verbosity from INFO to ERROR")
    parser.add_argument("--verbose", action="store_true", help="increase log verbosity from INFO to DEBUG")
    parser.add_argument("--debug", action="store_true", help="like --verbose but also include websockets logs")
    parser.add_argument("--config", type=str, help="path to configuration on disk")
    parser.add_argument("--logfile", type=str, help="path to logfile on disk (default is stdout)")
    parser.add_argument("--override", type=str, action="append", help='override a config parameter as "param:value"')

    args = parser.parse_args(args=argv)

    overrides = {} if not args.override else {token[0]: token[1] for token in [override.split(":") for override in args.override]}
    if args.logfile:
        overrides["logfile_path"] = args.logfile  # we want to expose this a little more explicitly in the argument list

    load_config(args.config, overrides)
    setup_logging(args.quiet, args.verbose, args.debug, config().logfile_path)

    server()


def run_demo(argv: List[str]) -> None:
    """Start the Apologies demo client."""

    parser = argparse.ArgumentParser(
        description="Start the apologies server demo client.",
        epilog="The client requires that you already have the server running.  "
        "By default, the client writes logs to stdout. If you prefer, you can "
        "specify the path to a logfile, and logs will be written there instead.  ",
    )

    parser.add_argument("--quiet", action="store_true", help="decrease log verbosity from INFO to ERROR")
    parser.add_argument("--verbose", action="store_true", help="increase log verbosity from INFO to DEBUG")
    parser.add_argument("--debug", action="store_true", help="like --verbose but also include websockets logs")
    parser.add_argument("--logfile", type=str, help="path to logfile on disk (default is stdout)")
    parser.add_argument("--host", type=str, default=DEFAULT_SERVER_HOST, help="host where the server is running")
    parser.add_argument("--port", type=int, default=DEFAULT_SERVER_PORT, help="port where the server is running on the host")

    args = parser.parse_args(args=argv)
    setup_logging(args.quiet, args.verbose, args.debug, args.logfile)
    demo(host=args.host, port=args.port)


def _example(argv: List[str]) -> List[str]:
    """Example method."""
    return argv[:]


def _lookup_method(method: str) -> Any:
    """Look up the method in this module with the passed-in name."""
    module = sys.modules[__name__]
    return getattr(module, "%s" % method)


def cli(script: str) -> Any:
    """
    Run the main routine for the named script.

    Args:
        script(str): Name of the script to execute
    """
    return _lookup_method(script)(argv=sys.argv[2:])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['apologiesserver']

package_data = \
{'': ['*']}

install_requires = \
['apologies>=0.1.19,<0.2.0',
 'asyncio-periodic>=2019.2,<2020.0',
 'ordered-set>=4.0.1,<5.0.0',
 'websockets>=8.1,<9.0']

setup_kwargs = {
    'name': 'apologiesserver',
    'version': '0.1.3',
    'description': 'Websocket server to interactively play the Apologies game',
    'long_description': '# Apologies Server\n\n![](https://img.shields.io/pypi/l/apologiesserver.svg)\n![](https://img.shields.io/pypi/wheel/apologiesserver.svg)\n![](https://img.shields.io/pypi/pyversions/apologiesserver.svg)\n![](https://github.com/pronovic/apologies-server/workflows/Test%20Suite/badge.svg)\n![](https://readthedocs.org/projects/apologies-server/badge/?version=latest&style=flat)\n\n[Apologies Server](https://github.com/pronovic/apologies-server) is a [Websocket](https://en.wikipedia.org/wiki/WebSocket) server interface used to interactively play a multi-player game using the [Apologies](https://github.com/pronovic/apologies) library.  The Apologies library implements a game similar to the [Sorry](https://en.wikipedia.org/wiki/Sorry!_(game)) board game.  See the [documentation](https://apologies-server.readthedocs.io/en/latest) for notes about the public interface and the event model.\n\n_Note:_ At present, the Apologies Server runs as a single stateful process that\nmaintains game state in memory.  It cannot be horizontally scaled, and there is\nno option for an external data store.  There is also only limited support for\nauthentication and authorization - basically, any player can register any\navailable handle.  We do enforce resource limits (open connections, registered\nusers, in-progress games) to limit the amount of damage abusive clients can do.\n',
    'author': 'Kenneth J. Pronovici',
    'author_email': 'pronovic@ieee.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pronovic/apologies-server',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
